/**
 * ddr_training_custom.c
 *
 * Copyright (c) 2009-2014, HiSilicon Technologies Co., Ltd.
 * All rights reserved.
 *
 * DDR training custom implement.
 */

#include <ddr_interface.h>
#include "ddr_training_impl.h"

#define CRG_REG_BASE 0xf8a20000U
#define PERI_CRG25	 0x64U

/**
 * ddr_cmd_prepare_copy
 * @void
 *
 * Do some prepare before copy code from DDR to SRAM.
 * Keep empty when nothing to do.
 */
void ddr_cmd_prepare_copy(void) { return; }

/**
 * ddr_cmd_site_save
 * @void
 *
 *
 */
void ddr_cmd_site_save(void)
{
	unsigned int ddrt_clk_reg;
	REG_WRITE(0x4db107df, 0xf8a220d8);

	/* turn on ddrt clock */
	ddrt_clk_reg = REG_READ(CRG_REG_BASE + PERI_CRG25);
	ddrt_clk_reg |= (1U << 6);	/* enable ddrt0 clock */
	REG_WRITE(ddrt_clk_reg, CRG_REG_BASE + PERI_CRG25);
	__asm__ __volatile__("nop");
	ddrt_clk_reg = REG_READ(CRG_REG_BASE + PERI_CRG25);
	ddrt_clk_reg &= ~(1U << 12);	/* disable ddrt0 soft reset */
	REG_WRITE(ddrt_clk_reg, CRG_REG_BASE + PERI_CRG25);
	return;
}

/**
 * ddr_cmd_site_restore
 * @void
 *
 *
 */
void ddr_cmd_site_restore(void)
{
	unsigned int ddrt_clk_reg;

	/* turn off ddrt clock */
	ddrt_clk_reg = REG_READ(CRG_REG_BASE + PERI_CRG25);
	ddrt_clk_reg |= (1U << 12);	/* enable ddrt0 soft reset */
	REG_WRITE(ddrt_clk_reg, CRG_REG_BASE + PERI_CRG25);
	__asm__ __volatile__("nop");
	ddrt_clk_reg = REG_READ(CRG_REG_BASE + PERI_CRG25);
	ddrt_clk_reg &= ~(1U << 6);	/* disable ddrt0 clock */
	REG_WRITE(ddrt_clk_reg, CRG_REG_BASE + PERI_CRG25);
}

#define  REG_BASE_MDDRC_SEC_4  0xff582000
#define  REG_BASE_MDDRC_SEC_5  0xff582400
#define  REG_BASE_MDDRC_QOS    0xff584000
void ddr_training_save_reg_custom(void *reg, unsigned int mask)
{
	int i = 0;
	struct tr_relate_reg *relate_reg = (struct tr_relate_reg *)reg;
	for (i = 0; i < DDR_PHY_NUM; i++) {
		if (ddr_training_phy_disable(i))
			continue;

		if (0 == i) {
			relate_reg->custom.sec_bypass[i] =
				REG_READ(REG_BASE_MDDRC_SEC_4 + 0x4);
			/* set sucure bypass */
			REG_WRITE(0x1, REG_BASE_MDDRC_SEC_4 + 0x4);
		} else {
			relate_reg->custom.sec_bypass[i] =
				REG_READ(REG_BASE_MDDRC_SEC_5 + 0x4);
			/* set sucure bypass */
			REG_WRITE(0x1, REG_BASE_MDDRC_SEC_5 + 0x4);
		}
	}
	relate_reg->custom.qosb_buff_byp = REG_READ(REG_BASE_MDDRC_QOS + 0x88);
	/* set qosbuf bypass */
	REG_WRITE(0x3, REG_BASE_MDDRC_QOS + 0x88);
}
void ddr_training_restore_reg_custom(void *reg)
{
	int i = 0;
	struct tr_relate_reg *relate_reg = (struct tr_relate_reg *)reg;
	for (i = 0; i < DDR_PHY_NUM; i++) {
		if (ddr_training_phy_disable(i))
			continue;

		if (0 == i)
			REG_WRITE(relate_reg->custom.sec_bypass[i],
				REG_BASE_MDDRC_SEC_4 + 0x4);
		else
			REG_WRITE(relate_reg->custom.sec_bypass[i],
				REG_BASE_MDDRC_SEC_5 + 0x4);
	}

	REG_WRITE(relate_reg->custom.qosb_buff_byp,
		REG_BASE_MDDRC_QOS + 0x88);
}
