@/******************************************************************************
@ *    Copyright (c) 2009-2012 by Hisi.
@ *    All rights reserved.
@ * ***
@ *    Create by Czyong. 2012-09-07
@ *
@******************************************************************************/

#include <config.h>

/******************************************************************************/
#define STACK_TOP               0xFFFF0C00
#define BOOT_RUN_START_ADDR     0xFFFF0C00
/******************************************************************************/

.globl _start
_start:
	b	reset /* MBR start */
	ldr	pc, _undefined_instruction
	ldr	pc, _software_interrupt
	ldr	pc, _prefetch_abort
	ldr	pc, _data_abort
	ldr	pc, _not_used
	ldr	pc, _irq
	ldr	pc, _fiq

_undefined_instruction: .word irq_undefined_instruction
_software_interrupt:    .word irq_software_interrupt
_prefetch_abort:        .word irq_prefetch_abort
_data_abort:            .word irq_data_abort
_not_used:              .word irq_not_used
_irq:                   .word irq_irq
_fiq:                   .word irq_firq
_pad1:                  .word 0x12345678

.=0x200+0x3C
_pad2:                  .word 0x12345678
.=0x40+0x200
__blank_zone_start:
	.fill 1024*8,1,0
__blank_zone_end:

.balignl 16,0xdeadbeef
__start:
	.word _start

.globl _blank_zone_start
_blank_zone_start:
	.word __blank_zone_start


.globl _blank_zone_end
_blank_zone_end:
	.word __blank_zone_end

_TEXT_BASE:
	.word TEXT_BASE

.globl _armboot_start
_armboot_start:
	.word _start

/*
 * These are defined in the board-specific linker script.
 */
.globl _bss_start
_bss_start:
	.word __bss_start

.globl _bss_end
_bss_end:
	.word _end

.globl _relocate
_relocate:
	.word relocate

_clr_remap_nand_entry:
	.word MEM_BASE_NAND + do_clr_remap - TEXT_BASE

_clr_remap_spi_entry:
	.word MEM_BASE_SPI + do_clr_remap - TEXT_BASE

_copy_abort_code:
	.word  copy_abort_code

/*
 * the actual reset code
 */
reset:
	/*
	 * set cpu alive flag
	*/
	ldr	r0, =0xF80000B4 @GEN13
	ldr	r1, =0x6C697665 @magic num, means cpu is alive
	str	r1, [r0]

	/*
	 *  delay wait mcu switch BUS finish
	 *  prevent mcu and A9 compete BUS conflict
	 */
	ldr	r0, =0x100
delay_wait_mcu:
	sub	r0, #1
	cmp	r0, #0
	bne	delay_wait_mcu
	/*
	 *  read and write system register REG_SC_GEN1
	 *  judge BUS mux whether or not OK
	 */
	ldr	r0, =REG_BASE_SCTL
	mov	r1, #0
wait_mux_finish:
	add	r1, r1, #1
	str	r1, [r0, #REG_SC_GEN1]
	ldr	r2, [r0, #REG_SC_GEN1]
	cmp	r1, r2
	bne	wait_mux_finish

bootstrap_check:
	/*
	 *  read system register REG_SC_GEN14
	 *  check if ziju flag
	 */
	ldr	r1, [r0, #REG_SC_GEN14]
	ldr	r2, =0x7a696a75          /* magic for "ziju" */
	cmp	r1, r2
	bne	normal_start_flow

bootstrap:
	mov	r1, sp                 /* save sp */
	str	r1, [r0, #REG_SC_GEN14]  /* clear ziju flag */

	/* init PLL/DDRC/pin mux/... */
	ldr	r0, _blank_zone_start
	ldr	r1, _TEXT_BASE
	sub	r0, r0, r1
	add	r0, r0, #RAM_START_ADRS
	mov	r1, #0x0                /* flags: 0->normal 1->pm */
	mov	r2, #INIT_REG_ITEM1
	bl	init_registers           /* init PLL/DDRC/... */

	ldr	r0, =REG_BASE_SCTL
	ldr	r1, [r0, #REG_SC_GEN14]

	mov	sp, r1                   /* restore sp */
	ldr	r1, [r0, #REG_SC_GEN15]
	mov	pc, r1                   /* return to bootrom */

normal_start_flow:

	/* set the cpu to SVC32 mode */
	mrs	r0, cpsr
	bic	r0, r0, #0x1f  /* r0 = ((~0x1F) & r0) */
	orr	r0, r0, #0xd3  /* r0 = (0xd3 | r0)    */
	msr	cpsr, r0

	mrrc 	p15,1,r0,r1,c15
	orr	r0, r0, #0x40
	mcrr 	p15,1,r0,r1,c15
	bl	flash_cache_all

main_cpu:
	/* init serial and printf a string. */
	bl	uart_early_init
	bl	msg_main_cpu_startup

check_boot_type:
	/* check_boot_type */
	ldr	r0, =REG_BASE_PERI_CTRL
	ldr	r0, [r0, #REG_START_MODE]
	mov	r6, r0, lsr#9             /* r6 = (r0 >> 9) */
	and	r6, #0x3
	
	/*
	 * REG_SC_GEN14: save start addr
	 * REG_SC_GEN15: save run flag
	 */
	ldr	r4, =REG_BASE_SCTL

	cmp	r6, #BOOT_FROM_NAND
	bne	check_from_spi
#ifdef CONFIG_GENERIC_NAND
	ldr	r0, =REG_BASE_NANDC
	ldr	r0, [r0, #HINFC504_BOOT_CFG]
	and	r0, #0x01
	cmp	r0, #0x01
	bne	check_from_nand

check_from_nand:
	ldr	r2, =MEM_BASE_NAND        /* 0xFE00_0000 */
	str	r2, [r4, #REG_SC_GEN14]
	mov	r2, #BOOT_FROM_NAND
	str	r2, [r4, #REG_SC_GEN15]
	b	check_start_mode
#endif /* CONFIG_GENERIC_NAND */

check_from_spi:
	cmp	r6, #BOOT_FROM_SPI
	bne	check_from_emmc
#ifdef CONFIG_GENERIC_SF
	ldr	r2, =MEM_BASE_SPI           /* 0xFA00_0000 */
	str	r2, [r4, #REG_SC_GEN14]
	mov	r2, #BOOT_FROM_SPI
	str	r2, [r4, #REG_SC_GEN15]
	b	check_start_mode
#endif /* CONFIG_GENERIC_SF */
check_from_emmc:
	cmp	r6, #BOOT_FROM_EMMC
#ifdef CONFIG_FSD
	cmpne	r6, #BOOT_FROM_SD
#endif
	bne	bad_start_type
#ifdef CONFIG_GENERIC_MMC
	ldr	r2, =(BOOT_RUN_START_ADDR + CONFIG_MMC_BOOT_ADDR)
	str	r2, [r4, #REG_SC_GEN14]
	str	r6, [r4, #REG_SC_GEN15]
	b	check_start_mode
#endif
bad_start_type:
	b	msg_bad_start_type

check_start_mode:

	/* Check if I need jump to flash */
	mov	r0, pc, lsr#28             /* r0 = (pc >> 28) */
	cmp	r0, #0xF
	bne	check_ddr_run

	cmp	r6, #BOOT_FROM_NAND
	ldreq	pc, _clr_remap_nand_entry

	cmp	r6, #BOOT_FROM_SPI
	ldreq	pc, _clr_remap_spi_entry

	/* boot from emmc, I need jump after ddr_init,not here!*/
	cmp	r6, #BOOT_FROM_EMMC
#ifdef CONFIG_FSD
	cmpne	r6, #BOOT_FROM_SD
#endif
	beq	do_clr_remap

	b	msg_bad_start_type  /*bug here*/

check_ddr_run:

	cmp	r0, #0xF
	beq	do_clr_remap
	mov	r2, #BOOT_FROM_DDR
	str	r2, [r4, #REG_SC_GEN15]

do_clr_remap:
	ldr	r0, [r4, #REG_SC_CTRL]  /* r4 = 0xF800_0000 */

	/* Set clear remap bit */
	orr	r0, #(1<<8)             /* r0 = (r0 | (1 << 8)) */
	str	r0, [r4, #REG_SC_CTRL]  /* p156 */
	ldr	sp, =STACK_TOP

	/* enable I-Cache now */
	mrc	p15, 0, r0, c1, c0, 0
	orr	r0, r0, #0x00001000     /* set bit 12 (I) I-Cache */
	mcr	p15, 0, r0, c1, c0, 0

	/* check if pmoc resume */
	ldr	r1, [r4, #REG_SC_GEN27]
	ldr	r2, =0x66031013          /* magic for pm */
	cmp	r1, r2
	bne	un_resume_start

resume_start:
	/* check if suspend in 64bit system */
	ldr	r4, =REG_BASE_SCTL
	ldr	r1, [r4, #REG_SC_GEN28]
	ldr	r2, =0x64646464
	cmp	r1, r2
	beq	resume64_1

	mov	r1, #0                   /* clear pm flag */
	str	r1, [r4, #REG_SC_GEN27]
resume64_1:
	/* disable mcu */
	ldr	r0, =0xf840f000
	ldr	r1, [r0]
	bic	r1, r1, #0x1
	str	r1, [r0]

	/* disable timer0 */
	ldr	r0, =REG_BASE_TIMER01             /* r0 = 0xF8002000 */
	mov	r1, #0
	str	r1, [r0, #REG_TIMER_CONTROL]

	/* disable WDG_RST */
	mov	r1, #0
	str	r1, [r4, #REG_SC_WDG_RST_CTRL]

	ldr	r0, _blank_zone_start
	ldr	r1, _TEXT_BASE
	sub	r0, r0, r1
	ldr	r1, [r4, #REG_SC_GEN14]
	add	r0, r0, r1
	mov	r1, #1                   /* flags: 0->normal 1->pm */
	mov	r2, #INIT_REG_ITEM1_ITEM2
	bl	init_registers

	/* Restore DDRPHY Register */
	/* DDRPHY0 copy reg 0xf840E100 to 0xFF58C210, 20*4 regs*/
	mov	r4, #4
	ldr	r1, =0xf840E100
	ldr	r3, =0xFF58C210

copy_loop_ddr0:
	sub	r4, r4, #1
	mov	r0, #20

restore_ddr_phy0_reg1:
	sub	r0, r0, #1
	ldr	r2, [r1], #4
	str	r2, [r3], #4
	cmp	r0, #0
	bne	restore_ddr_phy0_reg1

	add	r3, r3, #0x30 @ 20regs=0x50 then 0x50+0x30=0x80
	cmp	r4, #0
	bne	copy_loop_ddr0

	ldr	r0, =0xFF58C000
	
	ldr	r1, [r0, #0x70]
	orr	r1, r1, #0x100000
	str	r1, [r0, #0x70]
	
	bic	r1, r1, #0x100000
	str	r1, [r0, #0x70]
	
	mov	r1, #0x8000      
	str	r1, [r0, #0x04]
	
	mov	r1, #0
	str	r1, [r0, #0x04]
	
	/* DDRPHY1 copy reg 0xf840E300 to 0xFF58D210, 20*4 regs*/
	mov	r4, #4
	ldr	r1, =0xf840E300
	ldr	r3, =0xFF58D210

copy_loop_ddr1:
	sub	r4, r4, #1
	mov	r0, #20

restore_ddr_phy1_reg1:
	sub	r0, r0, #1
	ldr	r2, [r1], #4
	str	r2, [r3], #4
	cmp	r0, #0
	bne	restore_ddr_phy1_reg1

	add	r3, r3, #0x30 @ 20regs=0x50 then 0x50+0x30=0x80
	cmp	r4, #0
	bne	copy_loop_ddr1

	ldr	r0, =0xFF58D000
	
	ldr	r1, [r0, #0x70]
	orr	r1, r1, #0x100000
	str	r1, [r0, #0x70]
	bic	r1, r1, #0x100000
	str	r1, [r0, #0x70]
	
	mov	r1, #0x8000      
	str	r1, [r0, #0x04]
	
	mov	r1, #0
	str	r1, [r0, #0x04]	

	/* check if suspend in 64bit system */
	ldr	r4, =REG_BASE_SCTL
	ldr	r1, [r4, #REG_SC_GEN28]
	ldr	r2, =0x64646464
	cmp	r1, r2
	beq	resume64_2
	
	ldr	r4, =REG_BASE_SCTL
	ldr	r0, [r4, #REG_SC_GEN9] /* goto the kernel */

	mov	pc, r0
	
resume64_2:	
	/* prepare warm reset */
	//set 0xf8a208c0=0xfff0000
	ldr	r0, =0xf8a208c0
	ldr	r1, =0xfff0000
	str	r1, [r0]
	
	//set 0xf8a208c4 bit0=1 
	ldr	r0, =0xf8a208c4
	ldr	r1, [r0]
	orr	r1, r1, #0x1
	str	r1, [r0]
	
	/* warm reset cpu */
	mrc p15, 0, r2, c12, c0, 2
	orr r2, r2, #0x3
	mcr p15, 0, r2, c12, c0, 2
	wfi

	/* Check if I'm running in ddr */
un_resume_start:
	ldr	r0, [r4, #REG_SC_GEN15]
	cmp	r0, #BOOT_FROM_DDR
	beq	relocate

normal_init_item1:
	ldr	r0, _blank_zone_start
	ldr	r1, _TEXT_BASE
	sub	r0, r0, r1
	ldr	r1, [r4, #REG_SC_GEN14]
	add	r0, r0, r1
	mov	r1, #0x0                 /* flags: 0->normal 1->pm */
	mov	r2, #INIT_REG_ITEM1
	bl	init_registers
	b	normal_init_item2

relocate:                            /* relocate U-Boot to RAM */
	mov	r2, pc
	sub	r2, r2, #8       /* Get the ddr address of relocate label */
	ldr	r1, _armboot_start
	ldr	r0, _relocate
	sub	r1, r0, r1
	sub	r0, r2, r1       /* Get the ddr address where the fastboot loaded */
	str	r0, [r4, #REG_SC_GEN14]

normal_init_item2:
	/* in any case , we will init item2 here */
	ldr	r0, _blank_zone_start
	ldr	r1, _TEXT_BASE
	sub	r0, r0, r1
	ldr	r1, [r4, #REG_SC_GEN14]
	add	r0, r0, r1
#ifdef CONFIG_SHOW_REG_INFO
	bl	display_register_info
#endif /* CONFIG_SHOW_REG_INFO */
	mov	r1, #0x0              /* flags: 0->normal 1->pm */
	mov	r2, #INIT_REG_ITEM2   /* only init item2 */
	bl	init_registers

check_boot_mode:
	ldr	r4, =REG_BASE_SCTL
	ldr	r2, [r4, #REG_SC_GEN15]
	cmp	r2, #BOOT_FROM_EMMC
#ifdef CONFIG_FSD
	cmpne	r2, #BOOT_FROM_SD
#endif
	beq	emmc_boot
	b	copy_abort_code

emmc_boot:
#ifdef CONFIG_GENERIC_MMC
	ldr	r0, _TEXT_BASE
	ldr	r1, _bss_start
	sub	r1, r1, r0
	bl	emmc_boot_read
	b	jump_to_ddr
#endif

jump_to_ddr:
	ldr	r4, =REG_BASE_SCTL
	ldr	r0, _TEXT_BASE
	str	r0, [r4, #REG_SC_GEN14]
	ldr	pc, _copy_abort_code

copy_abort_code:

	ldr	r4, =REG_BASE_SCTL
	ldr	r0, [r4, #REG_SC_GEN14]
	ldr	r1, =0xFFFF0000
	mov	r2, #0x4000

	/* memcpy(r1, r0, r2) */
	bl	memcpy

copy_to_ddr:
	/* relocate Boot to DDR */
	ldr	r4, =REG_BASE_SCTL
	ldr	r0, [r4, #REG_SC_GEN14]

	/* now, r0 stores __reset offset from where we get started */
	ldr	r1, _TEXT_BASE   /* r1 stores where we will copy uboot to */

	/* compare source and target address, if equal no copy to target address */
	cmp	r0, r1
	beq	stack_setup

	ldr	r2, _armboot_start
	ldr	r3, _bss_start
	sub	r2, r3, r2      /* r2 <- size of armboot */
	/* memcpy(r1, r0, r2) */
	bl	memcpy

	/* Set up the stack */
stack_setup:
	ldr	r0, _TEXT_BASE                      @ upper 128 KiB: relocated uboot
	sub	r0, r0, #CONFIG_BOOTHEAD_GAP        @ boot head gap
	sub	r0, r0, #CONFIG_SYS_MALLOC_LEN      @ malloc area
	sub	r0, r0, #CONFIG_SYS_GBL_DATA_SIZE   @ bdinfo
	sub	sp, r0, #12              @ leave 3 words for abort-stack
	and	sp, sp, #~7              @ 8 byte alinged for (ldr/str)d

	/* Clear BSS (if any). Is below tx (watch load addr - need space) */
clear_bss:
	ldr	r0, _bss_start      @ find start of bss segment
	ldr	r1, _bss_end        @ stop here
	mov	r2, #0x00000000     @ clear value

clear_bss_loop:
	str	r2, [r0]        @ clear BSS location
	cmp	r0, r1          @ are we at the end yet
	add	r0, r0, #4      @ increment clear index pointer
	bne	clear_bss_loop  @ keep clearing till at end

	mov	r0, #0
	/* Instruction sync and invalidate I-cache */
	mcr	p15, 0, r0, c7, c5, 4
	mcr	p15, 0, r0, c7, c5, 0
	mcr	p15, 0, r0, c7, c5, 6

	ldr	pc, _start_armboot  @ jump to C code

_start_armboot: .word start_armboot

@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@
@	void memcpy(r1, r0, r2);
@
.align	2
memcpy:
	add	r2, r0, r2
memcpy_loop:
	ldmia	r0!, {r3 - r10}
	stmia	r1!, {r3 - r10}
	cmp	r0, r2
	ble	memcpy_loop
	mov	pc, lr

@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@
@  corrupt: r0, r1, r2, r3
@
.align	2
flash_cache_all:

	/* disable MMU stuff and caches */
	mrc	p15, 0, r0, c1, c0, 0
	orr	r0, r0, #0x00002000      /* clear bits 13 (--V-)   */
	bic	r0, r0, #0x00000007      /* clear bits 2:0 (-CAM)  */
	orr	r0, r0, #0x00000002      /* set bit 1 (--A-) Align */
	orr	r0, r0, #0x00000800      /* set bit 12 (Z---) BTB  */
	mcr	p15, 0, r0, c1, c0, 0

	/*
	 * Invalidate L1 I/D
	 */
	mov	r0, #0                   /* set up for MCR */
	mcr	p15, 0, r0, c8, c7, 0    /* invalidate TLBs */
	mcr	p15, 0, r0, c7, c5, 0    /* invalidate icache */
#if 0
	/* Invalidate L1 D-cache */
	mcr	p15, 2, r0, c0, c0, 0    /* select L1 data cache */
	mrc	p15, 1, r3, c0, c0, 0    /* Read Current Cache Size Identification Register */
	ldr	r1, =0x1ff
	and	r3, r1, r3, LSR #13      /* r3 = (number of sets -1) */
	mov	r0, #0
way_loop:
	mov	r1, #0                  /* r1->set counter */
line_loop:
	mov	r2, r0, LSL #30
	orr	r2, r1, LSL #5           /* r2->set/way cache-op format */
	mcr     p15, 0, r2, c7, c6, 2    /* Invalidate line described by r2 */
	add	r1, r1, #1               /* Increment set counter */
	cmp	r1, r3                   /* Check if the last set is reached... */
	ble	line_loop                /* if not, continue the set_loop */
	add	r0, r0, #1               /* else, Increment way counter */
	cmp	r0, #4                   /* Check if the last way is reached... */
	blt	way_loop                 /* if not, continue the way_loop */
#endif
	mov	pc, lr
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

.align	2
msg_main_cpu_startup:
	mov	r5, lr
	add	r0, pc, #4
	bl	uart_early_puts
	mov	pc, r5
L10:
	.ascii "\r\n\r\nSystem startup\r\n\0"

@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

.align	2
msg_bad_start_type:
	add	r0, pc, #4
	bl	uart_early_puts
	b	reset
L30:
	.ascii "* Not support current start mode\r\n\0"
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#ifdef CONFIG_SHOW_REG_INFO
@
@  int display_regv110_info(unsigned int base);
@
.align	2
.global	display_regv110_info
.type	display_regv110_info, %function
display_regv110_info:
	push	{r0, r4, lr}
	mov	r4, r0
	adr	r0, display_regv110_info_str0
	bl	uart_early_puts
	mov	r0, r4
	bl	uart_early_puts
	adr	r0, display_regv110_info_str1
	bl	uart_early_puts
	b	display_regv110_info_L0
display_regv110_info_str0:
	.asciz "\r\nReg Version:	"
display_regv110_info_str1:
	.asciz "\r\nReg Time:	"

.align 2
display_regv110_info_L0:
	add	r0, r4, #0x40
	bl	uart_early_puts
	add	r0, pc, #4
	bl	uart_early_puts
	b	display_regv110_info_L1
display_regv110_info_str2:
	.ascii "\r\nReg Name:	\0"

.align 2
display_regv110_info_L1:
	add	r0, r4, #0x52
	bl	uart_early_puts
	pop	{r0, r4, pc}

@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@
@  int display_regunknow_info(unsigned int base);
@
.align	2
.global	display_regunknow_info
.type	display_regunknow_info, %function
display_regunknow_info:
	push	{r0, lr}
	add	r0, pc, #4
	bl	uart_early_puts
	b	display_regunknow_info_L0
display_regunknow_info_str0:
	.ascii "Unknow register table\r\n\0"
.align 2
display_regunknow_info_L0:
	pop	{r0, pc}

#endif /* CONFIG_SHOW_REG_INFO */
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

.align	2
.globl _text_end
_text_end: .word __text_end

/* exception handlers */
#define _MODE_SVC               0x13
#define _REG_SPACE             72

_irq_stack_start:
	.word TEXT_BASE

.macro irq_save_to_stack
	ldr	sp, _irq_stack_start
	str	lr, [sp, #-4]
	mrs	lr, spsr
	str	lr, [sp, #-8]
	mov	lr, #_MODE_SVC
	msr	spsr, lr
	mov	lr, pc
	movs	pc, lr
.endm

.macro irq_save_regs_to_stack
	ldr	sp, _irq_stack_start
	sub	sp, sp, #_REG_SPACE
	stm	sp, {r0 - r12}
	ldr	r2, _irq_stack_start
	sub	r2, r2, #8
	ldmia	r2, {r2 - r3}          /* r2 = lr; r3 = spsr */
	mov	r0, sp                 /* r0 = register store start address */
	add	r5, sp, #52
	mov	r1, lr
	stm	r5, {r0 - r3}
.endm

.macro get_bad_stack_swi
	sub	sp, sp, #4
	str	r0, [sp]
	ldr	r0, _irq_stack_start
	str	lr, [r0]
	mrs	r0, spsr
	str	lr, [r0, #4]
	ldr	r0, [sp]
	add	sp, sp, #4
.endm

.align	5
irq_undefined_instruction:
	irq_save_to_stack
	irq_save_regs_to_stack
	stmdb	sp!, {r0 - r3}
	add	r0, pc, #8
	bl	uart_early_puts
	ldmia	sp!, {r0 - r3}
	bl	do_undefined_instruction
str_undefined_instruction:
	.ascii "\r\n* irq: undefined instruction\r\n\0"

.align	5
irq_software_interrupt:
	get_bad_stack_swi
	irq_save_regs_to_stack
	stmdb	sp!, {r0 - r3}
	add	r0, pc, #8
	bl	uart_early_puts
	ldmia	sp!, {r0 - r3}
	bl	do_software_interrupt
str_software_interrupt:
	.ascii "\r\n* irq: software interrupt\r\n\0"

.align	5
irq_prefetch_abort:
	irq_save_to_stack
	irq_save_regs_to_stack
	stmdb	sp!, {r0 - r3}
	add	r0, pc, #8
	bl	uart_early_puts
	ldmia	sp!, {r0 - r3}
	bl	do_prefetch_abort
str_prefetch_abort:
	.ascii "\r\n* irq: prefetch abort\r\n\0"

.align	5
irq_data_abort:
	irq_save_to_stack
	irq_save_regs_to_stack
	stmdb	sp!, {r0 - r3}
	add	r0, pc, #8
	bl	uart_early_puts
	ldmia	sp!, {r0 - r3}
	bl	do_data_abort
str_data_abort:
	.ascii "\r\n* irq: data abort\r\n\0"

.align	5
irq_not_used:
	irq_save_to_stack
	irq_save_regs_to_stack
	stmdb	sp!, {r0 - r3}
	add	r0, pc, #8
	bl	uart_early_puts
	ldmia	sp!, {r0 - r3}
	bl	do_not_used
str_not_used:
	.ascii "\r\n* irq: not used\r\n\0"

.align	5
irq_irq:
	irq_save_to_stack
	irq_save_regs_to_stack
	stmdb	sp!, {r0 - r3}
	add	r0, pc, #8
	bl	uart_early_puts
	ldmia	sp!, {r0 - r3}
	bl	do_irq
str_irq:
	.ascii "\r\n* irq: interrupt request\r\n\0"

.align	5
irq_firq:
	irq_save_to_stack
	irq_save_regs_to_stack
	stmdb	sp!, {r0 - r3}
	add	r0, pc, #8
	bl	uart_early_puts
	ldmia	sp!, {r0 - r3}
	bl	do_fiq
str_fip:
	.ascii "\r\n* irq: fast interrupt request\r\n\0"

