/******************************************************************************
*    Copyright (c) 2009-2014 by Hisi.
*    All rights reserved.
* ***
*    Create by zengzhongliang. 2014-08-29
*
******************************************************************************/
#include <common.h>

/* bootm arm64 flag */
int g_arm64_switch = 0;

int do_arm64_booting(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	volatile u32 *cpu_addr_reg = (u32 *)ARM64_CPU_ADDR;
	volatile u32 *cpu_ctrl_reg = (u32 *)ARM64_CPU_CTRL;

	extern char arm64_boot_entry[], arm64_boot_end[];
	extern int do_bootm(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);

	if (argc != 4) {
		printf("arc=%d\n", argc);
		cmd_usage(cmdtp);
		return 1;
	}

	/* arm64 boot image prepare */
	memcpy((void *)ARM64_WM_RST_VEC, (void *)arm64_boot_entry, arm64_boot_end - arm64_boot_entry);
	*cpu_addr_reg = ARM64_WM_RST_VEC;	/* config warm reset vector */
	*cpu_ctrl_reg = *cpu_ctrl_reg | 1;	/* enable 64bit instruction mode */

	g_arm64_switch = 1;
	do_bootm(cmdtp, flag, argc, argv);

	return 0;
}

int do_arm64_go(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong	addr, rc;

	volatile u32 *cpu_addr_reg = (u32 *)ARM64_CPU_ADDR;
	volatile u32 *cpu_ctrl_reg = (u32 *)ARM64_CPU_CTRL;

	extern char arm64_boot_entry[], arm64_boot_end[];
	extern int do_go (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);

	if (argc < 2) {
		cmd_usage(cmdtp);
		return 1;
	}

	addr = simple_strtoul(argv[1], NULL, 16);

	*cpu_addr_reg = addr;
	*cpu_ctrl_reg = *cpu_ctrl_reg | 1;

	/* warm reseting */
	__asm__ __volatile__("isb\n"
			"dsb\n"
			"mrc p15, 0, r2, c12, c0, 2\n"
			"orr r2, r2, #0x3\n"
			"mcr p15, 0, r2, c12, c0, 2\n"
			"isb\n"
			"wfi");

	return 0;
}

U_BOOT_CMD(
	go64, CONFIG_SYS_MAXARGS, 1,	do_arm64_go,
	"start application at address 'addr'",
	"addr [arg ...]\n    - start application at address 'addr'\n"
	"      passing 'arg' as arguments"
);

U_BOOT_CMD(
	boot64, CONFIG_SYS_MAXARGS, 1,	do_arm64_booting,
	"boot application image from memory in arm64 mode",
	"example: boot64 kernel_addr - fdt_addr\n"
);
