/*
 * (C) Copyright 2001
 * Denis Peter, MPL AG Switzerland
 *
 * Most of this source has been derived from the Linux USB
 * project.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 */

#include <common.h>
#include <command.h>
#include <nand.h>
#include <nand_logif.h>
#include <spiflash_logif.h>
#include <emmc_logif.h>

typedef unsigned char           HI_U8;
typedef unsigned char           HI_UCHAR;
typedef unsigned short          HI_U16;
typedef unsigned int            HI_U32;

typedef signed char             HI_S8;
typedef short                   HI_S16;
typedef int                     HI_S32;

#ifndef _M_IX86
typedef unsigned long long      HI_U64;
typedef long long               HI_S64;
#else
typedef __int64                 HI_U64;
typedef __int64                 HI_S64;
#endif

typedef char                    HI_CHAR;
typedef char*                   HI_PCHAR;

typedef float                   HI_FLOAT;
typedef double                  HI_DOUBLE;
/*typedef void                    HI_VOID;*/
#define HI_VOID                void

typedef unsigned long           HI_SIZE_T;
typedef unsigned long           HI_LENGTH_T;

typedef HI_U32                  HI_HANDLE;

typedef enum hiHI_FLASH_TYPE_E
{
    HI_FLASH_TYPE_SPI_0,
    HI_FLASH_TYPE_NAND_0,
    HI_FLASH_TYPE_EMMC_0,

    HI_FLASH_TYPE_BUTT
} HI_FLASH_TYPE_E;

#define  HI_FLASH_RW_FLAG_RAW           0x0   /* read/write without oob, such as write kernel/uboot/ubi/cramfs.. */
#define  HI_FLASH_RW_FLAG_WITH_OOB      0x1   /* read/write nand with oob, such as write yaffs2 filesystem image */
#define  HI_FLASH_RW_FLAG_ERASE_FIRST   0x2   /* erase flash before write */
#define  FASTBOOT_MISC_NAME "misc"
#define  FASTBOOT_MISC_OFFSET 0
#define  ERR_FALSH_TRACE(fmt, arg...)   //printf( "%s,%d: " fmt , __FUNCTION__, __LINE__, ## arg)
#define  INFO_FLASH_TRACE(fmt, arg...)   //printf( "%s,%d: " fmt , __FUNCTION__, __LINE__, ## arg)

typedef struct hiHI_FLASH_HANDLE_S
{
    HI_FLASH_TYPE_E    type;
    union
    {
        nand_logic_t      *nand;
        spiflash_logic_t  *spiflash;
        emmc_logic_t      *emmc;
    } u;
} HI_FLASH_HANDLE_S;

typedef struct hiBOOTLOADER_MESSAGE {
    char command[32];
    char status[32];
    char recovery[1024];
}BOOTLOADER_MESSAGE;

#define SZ_1KB    (1024)

#define HI_NULL          0L
#define HI_NULL_PTR      0L

#define HI_SUCCESS       0
#define HI_FAILURE       (-1)

#define HI_INVALID_HANDLE (0xffffffff)

#define HI_FLASH_HANDLE_TO_STRUCT(handle)    ((HI_FLASH_HANDLE_S *)(handle))
#define HI_FLASH_STRUCT_TO_HANDLE(handle)    ((HI_HANDLE)(handle))

HI_S32 Adpt_emmc_write(emmc_logic_t *pstEmmcLog,
                          HI_U64 u64Offset,
                          HI_U32 u32Len,
                          HI_UCHAR *pucBuf)
{
    HI_S32 s32Ret;
    HI_U32 u32BlkSize;
    HI_U32 u32BlkOffset;
    HI_U32 u32WriteLen;
    HI_UCHAR *pucTmpBuff;

    if( NULL == pstEmmcLog
       || NULL == pucBuf)
    {
        ERR_FALSH_TRACE("Pointer is null.");
        return HI_FAILURE;
    }

    u32BlkSize = pstEmmcLog->blocksize * 16;
    u32BlkOffset = u32Len & (u32BlkSize - 1);
    INFO_FLASH_TRACE("emmc u64Offset is %u\n", u64Offset);
    INFO_FLASH_TRACE("emmc u32Len is %u\n", u32Len);
    u32WriteLen = u32Len - u32BlkOffset;
    INFO_FLASH_TRACE("emmc write length is %u\n", u32WriteLen);
    INFO_FLASH_TRACE("emmc u32BlkOffset is %u\n", u32BlkOffset);
    if(u32WriteLen)
    {
        INFO_FLASH_TRACE("emmc write into is %s\n", pucBuf);
        s32Ret = emmc_logic_write(pstEmmcLog, u64Offset, u32WriteLen, pucBuf);
        if( HI_SUCCESS != s32Ret)
        {
            return s32Ret;
        }
    }

    if( u32BlkOffset )
    {
        if( NULL ==(pucTmpBuff = (HI_UCHAR*)malloc(u32BlkSize)))
        {
            ERR_FALSH_TRACE("Failed to allocate memory.");
            return HI_FAILURE;
        }

        memset(pucTmpBuff, 0xFF, u32BlkSize);
        memcpy(pucTmpBuff, pucBuf + u32WriteLen, u32BlkOffset);
        INFO_FLASH_TRACE("emmc write into is %s\n", pucBuf);
        INFO_FLASH_TRACE("emmc write final into is %s\n", pucTmpBuff);
        s32Ret = emmc_logic_write(pstEmmcLog, u64Offset + u32WriteLen, u32BlkSize, pucTmpBuff);
        if( HI_SUCCESS != s32Ret)
        {
            free(pucTmpBuff);
            return s32Ret;
        }
        free(pucTmpBuff);
    }
    return u32Len;
}

HI_S32 Flash_Write(HI_HANDLE hFlash, HI_U64 Offset,
                      HI_U8 *pBuf, HI_U32 Len, HI_U32 Flags)
{
    HI_FLASH_HANDLE_S *flash_handle = HI_FLASH_HANDLE_TO_STRUCT(hFlash);

    return Adpt_emmc_write(flash_handle->u.emmc, Offset, Len, pBuf);
}

HI_HANDLE Flash_OpenByTypeAndName(HI_FLASH_TYPE_E enFlashType, HI_CHAR *pPartitionName)
{
    HI_FLASH_HANDLE_S *flash_handle;
    HI_U64 u64Address = 0;
    HI_U64 u64Len = 0;
    char *ptr;
    struct mtd_info_ex *mtdinfo;
    const char *bootargs_str = (const char *)getenv("bootargs");
    char media_name[20];

    INFO_FLASH_TRACE("enFlashType = %d,pPartitionName =%s\n",enFlashType, pPartitionName);
    if((pPartitionName == HI_NULL) || (enFlashType == HI_FLASH_TYPE_BUTT) || (bootargs_str == HI_NULL))
    {
        ERR_FALSH_TRACE("enFlashType = %d,pPartitionName =%s\n",enFlashType,pPartitionName);
        return HI_INVALID_HANDLE;
    }

    if ((flash_handle = (HI_FLASH_HANDLE_S *)malloc(sizeof(HI_FLASH_HANDLE_S))) == NULL)
    {
        ERR_FALSH_TRACE("not enough memory.\n");
        return HI_INVALID_HANDLE;
    }

    flash_handle->type = enFlashType;

    if ((ptr = strstr(bootargs_str, "blkdevparts=")) == NULL)
    {
        ERR_FALSH_TRACE("Cannot find bootargs\n");
        free(flash_handle);
        return HI_INVALID_HANDLE;
    }

    ptr += strlen("blkdevparts=");
    memset(media_name, 0 , 20);
    memcpy(media_name, "mmcblk0", strlen("mmcblk0"));
    if (find_flash_part(ptr, media_name, pPartitionName, &u64Address, &u64Len) == 0)
    {
        ERR_FALSH_TRACE("Cannot find partition: %s\n", pPartitionName);
        free(flash_handle);
        return HI_INVALID_HANDLE;
    }

    flash_handle->u.emmc = emmc_logic_open(u64Address, u64Len);
    if( NULL == flash_handle->u.emmc)
    {
        free(flash_handle);
        return HI_INVALID_HANDLE;
    }

    return HI_FLASH_STRUCT_TO_HANDLE(flash_handle);
}

HI_HANDLE Flash_OpenByName(HI_CHAR *pPartitionName)
{
    HI_S32 index = 0;
    HI_HANDLE hFlashHandle = HI_INVALID_HANDLE;

    hFlashHandle = Flash_OpenByTypeAndName(HI_FLASH_TYPE_EMMC_0,pPartitionName);
    if(HI_INVALID_HANDLE != hFlashHandle)
    {
      return hFlashHandle;
    }
    return HI_INVALID_HANDLE;
}

void set_misc_db(BOOTLOADER_MESSAGE boot, int len)
{
    HI_HANDLE   flashhandle;
    flashhandle = Flash_OpenByName(FASTBOOT_MISC_NAME);
    HI_U8 buf[len / sizeof(HI_U8)];
    memset(&buf, 0, sizeof(buf));
    memcpy(buf, &boot, sizeof(BOOTLOADER_MESSAGE));
    HI_S32 ret;
    if ((0 == flashhandle) || (HI_INVALID_HANDLE == flashhandle)) {
        printf("HI_Flash_Open partion %s error\n", FASTBOOT_MISC_NAME);
    } else {
        ret = Flash_Write(flashhandle, FASTBOOT_MISC_OFFSET, buf, len, HI_FLASH_RW_FLAG_RAW);
        printf("set_misc_db ret = %d\n", ret);
        if(ret == HI_FAILURE)
            printf("set_misc_db failed\n");
    }
}

int enter_recovery (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
    int exitRecovery = -1;
    const int length = 6 * SZ_1KB * sizeof(unsigned char);
    BOOTLOADER_MESSAGE boot;
    memset(&boot, 0, sizeof(boot));
    if (argc == 1) {
        strncpy(boot.command, "boot-recovery", sizeof(boot.command));
        strncpy(boot.recovery, "recovery\n", sizeof(boot.recovery));
        set_misc_db(boot, length);
        exitRecovery = 0;
    }
    else if (argc == 2) {
        if (strcmp(argv[1], "no") == 0) {
            set_misc_db(boot, length);
            exitRecovery = 1;
        }
        else if (strcmp(argv[1], "yes") == 0) {
            strncpy(boot.command, "boot-recovery", sizeof(boot.command));
            strncpy(boot.recovery, "recovery\n", sizeof(boot.recovery));
            set_misc_db(boot, length);
            exitRecovery = 0;
        }
    }
    if (exitRecovery == -1) {
       cmd_usage(cmdtp);
       return 0;
    }
    printf("%s Recovery ! Now Reboot First...\n", exitRecovery ? "Exit" : "Enter");
    do_reset();
    return 0;
}

#if defined(CONFIG_CMD_ENTER_RECOVERY)
U_BOOT_CMD(
    recovery, 2, 0,    enter_recovery,
    "enter recovery system directly from fastboot",
    "[yes|no]"
);
#endif
