#include <common.h>
#include <malloc.h>
#include <mmc.h>
#include <dwmmc.h>
#include <asm/arch/platform.h>

extern void udelay(unsigned long usec);
extern int mmc_send_ext_csd(struct mmc *mmc, u8 *ext_csd);
extern int mmc_switch(struct mmc *mmc, u8 set, u8 index, u8 value);

#define UNSTUFF_BITS(resp,start,size)                   \
    ({                              \
        const int __size = size;                \
        const u32 __mask = (__size < 32 ? 1 << __size : 0) - 1; \
        const int __off = 3 - ((start) / 32);           \
        const int __shft = (start) & 31;            \
        u32 __res;                      \
                                    \
        __res = resp[__off] >> __shft;              \
        if (__size + __shft > 32)               \
            __res |= resp[__off-1] << ((32 - __shft) % 32); \
        __res & __mask;                     \
    })

struct card_cid {
	unsigned int        manfid;
	char            prod_name[8];
	unsigned char       prv;
	unsigned int        serial;
	unsigned short      oemid;
	unsigned short      year;
	unsigned char       hwrev;
	unsigned char       fwrev;
	unsigned char       month;
};

int print_mmc_reg(int dev_num, int show_ext_csd)
{
	struct mmc *mmc;
	unsigned int tmp;
	char ext_csd[512];

	mmc = find_mmc_device(dev_num);
	if (!mmc) {
		printf("No MCI found!!!\n");
		return -1;
	}

	printf("OCR register: %08x\n", mmc->ocr);
	printf("CID register: %08x %08x %08x %08x\n",
	       mmc->cid[0], mmc->cid[1], mmc->cid[2], mmc->cid[3]);
	printf("CSD register: %08x %08x %08x %08x\n",
	       mmc->csd[0], mmc->csd[1], mmc->csd[2], mmc->csd[3]);
	printf("RCA register: %08x\n", mmc->rca);
	if (show_ext_csd) {
		int err = mmc_send_ext_csd(mmc, ext_csd);
		if (err) {
			printf("Check est_csd error!\n");
			return -1;
		}
		printf("Extended CSD register:\n");
		for (tmp = 0; tmp < 512; tmp += 8)
			printf("%03d: %02x %02x %02x %02x %02x %02x %02x %02x\n",
			       tmp,
			       ext_csd[tmp],
			       ext_csd[tmp+1],
			       ext_csd[tmp+2],
			       ext_csd[tmp+3],
			       ext_csd[tmp+4],
			       ext_csd[tmp+5],
			       ext_csd[tmp+6],
			       ext_csd[tmp+7]);
	}
	printf("\n");
	return 0;
}

void print_mmcinfo(struct mmc *mmc)
{
	u32 *resp = mmc->cid;
	struct card_cid card_cid;
	struct dwmci_host *host = mmc->priv;
	memset(&card_cid, 0, sizeof(struct card_cid));

	card_cid.manfid    = UNSTUFF_BITS(resp, 120, 8);
	card_cid.oemid     = UNSTUFF_BITS(resp, 104, 16);
	card_cid.prod_name[0]  = UNSTUFF_BITS(resp, 96, 8);
	card_cid.prod_name[1]  = UNSTUFF_BITS(resp, 88, 8);
	card_cid.prod_name[2]  = UNSTUFF_BITS(resp, 80, 8);
	card_cid.prod_name[3]  = UNSTUFF_BITS(resp, 72, 8);
	card_cid.prod_name[4]  = UNSTUFF_BITS(resp, 64, 8);
	card_cid.prod_name[5]  = UNSTUFF_BITS(resp, 56, 8);
	card_cid.prv       = UNSTUFF_BITS(resp, 48, 8);
	card_cid.serial    = UNSTUFF_BITS(resp, 16, 32);
	card_cid.month     = UNSTUFF_BITS(resp, 12, 4);
	card_cid.year      = UNSTUFF_BITS(resp, 8, 4) + 2013;

	printf("MMC/SD Card:\n");
	printf("    MID:         0x%06x\n", card_cid.manfid);
	printf("    OID:         0x%04x\n", card_cid.oemid);
	printf("    PSN:         0x%08x\n", card_cid.serial);
	printf("    DATA:        %02d/%04d\n",
			card_cid.month, card_cid.year);
	printf("    Name:        \"%s\"\n", card_cid.prod_name);
	printf("    Chip Type:   %s\n", IS_SD(mmc) ? "SD" : "MMC");

	switch (mmc->version) {
		case SD_VERSION_1_0:
			printf("    SD Version:  1.0\n");
			break;
		case SD_VERSION_1_10:
			printf("    SD Version:  1.10\n");
			break;
		case SD_VERSION_2:
			printf("    SD Version:  2.0\n");
			break;
		case SD_VERSION_3:
			printf("    SD Version:  3.0\n");
			break;
		case MMC_VERSION_4_1:
			printf("    MMC Version: 4.1\n");
			break;
		case MMC_VERSION_4_2:
			printf("    MMC Version: 4.2\n");
			break;
		case MMC_VERSION_4_3:
			printf("    MMC Version: 4.3\n");
			break;
		case MMC_VERSION_4_41:
			printf("    MMC Version: 4.41\n");
			break;
		case MMC_VERSION_4_5:
			printf("    MMC Version: 4.5\n");
			break;
		case MMC_VERSION_5_0:
			printf("    MMC Version: 5.0\n");
			break;
		default:
			printf("    Version:     UNKNOWN\n");
			break;
	}

	switch (mmc->mode) {
		case MMC_MODE_HS:
			printf("    Timing:      HS\n");
			break;
		case MMC_MODE_HS_52MHz:
			printf("    Timing:      HS_52MHz\n");
			break;
		case MMC_MODE_DDR50:
			printf("    Timing:      DDR50\n");
			break;
	}

	printf("    Host Clk:    %uHz\n", host->bus_hz);
	printf("    Host Div:    %d\n", dwmci_readl(host, DWMCI_CLKDIV));
	printf("    Bus Width:   %dbit\n", mmc->bus_width);
	printf("    Chip Size:   %Lu Bytes (%s)\n",
	       mmc->capacity,
	       mmc->high_capacity ? "High Capacity" : "Low Capacity");
	printf("    Read Block:  %d Bytes\n", mmc->read_bl_len);
	printf("    Write Block: %d Bytes\n", mmc->write_bl_len);
	printf("    Boot Addr:   %d\n", CONFIG_MMC_BOOT_ADDR);
}

static void hi_mci_sys_init(void)
{
	unsigned int tmp_reg;

	/* SDIO clock phase */
	tmp_reg = readl(REG_BASE_CRG + REG_PERI_CRG40);
	tmp_reg &= ~SDIO1_CLK_SEL_MASK;
	tmp_reg &= ~SDIO1_DRV_PS_SEL_MASK;
	tmp_reg &= ~SDIO1_SAP_PS_SEL_MASK;
	tmp_reg |= SDIO1_CLK_SEL_50M | SDIO1_DRV_PS_SEL_135 | SDIO1_SAP_PS_SEL_45;
	writel(tmp_reg, REG_BASE_CRG + REG_PERI_CRG40);

	/* SDIO soft reset */
	tmp_reg |= SDIO1_SRST_REQ;
	writel(tmp_reg, REG_BASE_CRG + REG_PERI_CRG40);
	udelay(1000);
	tmp_reg &= ~SDIO1_SRST_REQ;
	tmp_reg |= SDIO1_CKEN | SDIO1_BUS_CKEN;
	writel(tmp_reg, REG_BASE_CRG + REG_PERI_CRG40);

	udelay(1000);
}

void check_ext_csd(struct mmc *mmc)
{
	u8 ext_csd[512];
	unsigned int tmp, tmp1, boot_part, boot_bus_width;
	int err = 0;

	/* check bootcfg on est_csd */
	err = mmc_send_ext_csd(mmc, ext_csd);
	if (err) {
		printf("Check est_csd error!");
		return;
	}

	if (get_ca_vendor())
		boot_part = 0x1;
	else
		boot_part = 0x7;

#if defined(CHIP_TYPE_hi3751v600) || defined(CHIP_TYPE_hi3751lv500)
	if (readl(REG_BASE_PERI_CTRL + REG_START_MODE) & NF_BOOTBW_MASK)
		boot_bus_width = 0x2; /* 8bits */
	else
		boot_bus_width = 0x1; /* 4bits */
#elif defined(CHIP_TYPE_hi3751v500)
	boot_bus_width = 0x2; /* 8bits */
#endif

	/*
	 * Architecture support reset pin
	 */
	tmp = ext_csd[EXT_CSD_RST_N_FUNCTION];
	tmp1 = ext_csd[EXT_CSD_REV];
	if ((tmp1 >= 5)
			&& ((tmp & EXT_CSD_RST_N_EN_MASK) != EXT_CSD_RST_N_ENABLED)) {
		err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL,
				EXT_CSD_RST_N_FUNCTION,
				EXT_CSD_RST_N_ENABLED);
		if (err) {
			printf("Set EXT_CSD_RST_N_FUNCTION error!");
			return;
		}
	}

	tmp = ext_csd[EXT_CSD_WR_REL_PARAM];
	tmp1 = ext_csd[EXT_CSD_REV];
	if ((tmp1 >= 5)
			&& ((tmp & EXT_CSD_WR_REL_PARAM_HCR) == EXT_CSD_WR_REL_PARAM_HCR)) {
		tmp1 = ext_csd[EXT_CSD_WR_REL_SET];
		if (tmp1 != EXT_CSD_WR_REL_VALUE) {
			err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL,
					EXT_CSD_WR_REL_SET,
					EXT_CSD_WR_REL_VALUE);
			if (err) {
				printf("Set EXT_CSD_WR_REL_SET error!");
				return;
			}
		}
	}

	tmp = ext_csd[EXT_CSD_PART_CONF] >> 0x3 & 0xF;
	tmp1 = ext_csd[EXT_CSD_BOOT_BUS_WIDTH] & 0x3;
	if (tmp == boot_part && tmp1 == boot_bus_width)
		return;

	if (tmp != boot_part) {
		printf("Set EXT_CSD_PART_CONF!\n");
		tmp = ext_csd[EXT_CSD_PART_CONF];
		tmp &= (~(0xF << 3));
		tmp |= (boot_part << 3);
		err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL,
				EXT_CSD_PART_CONF, tmp);
		if (err) {
			printf("Set EXT_CSD_PART_CONF error!");
			return;
		}
	}

	if (tmp1 != boot_bus_width) {
		printf("Set EXT_CSD_BOOT_BUS_WIDTH to 0x%X!\n", boot_bus_width);
		tmp1 = ext_csd[EXT_CSD_BOOT_BUS_WIDTH];
		tmp1 &= (~0x3);
		tmp1 |= boot_bus_width;
		err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL,
				EXT_CSD_BOOT_BUS_WIDTH, tmp1);
		if (err) {
			printf("set EXT_CSD_BOOT_BUS_WIDTH error!");
			return;
		}
	}

	err = mmc_send_ext_csd(mmc, ext_csd);
	if (err) {
		printf("Check est_csd error!");
		return;
	}

	tmp = ext_csd[EXT_CSD_PART_CONF] >> 0x3 & 0xF;
	tmp1 = ext_csd[EXT_CSD_BOOT_BUS_WIDTH] & 0x3;

	if (tmp == boot_part && tmp1 == boot_bus_width)
		printf("EXT_CSD CONFIG OK!\n");
	else
		printf("EXT_CSD CONFIG Fail!\n");
}

int s5_mmc_init(void)
{
	struct dwmci_host *host = NULL;
	unsigned int tmp_reg;
	int err = 0;

	host = malloc(sizeof(struct dwmci_host));
	if (!host) {
		printf("dwmci_host malloc fail!\n");
		return 1;
	}

	memset(host, 0, sizeof(struct dwmci_host));
	host->name = "Synopsys Mobile storage";
	host->ioaddr = (void *)REG_BASE_MCI;
	host->buswidth = 8;
	host->dev_index = 0;
	host->fifo_depth = 256;
	host->caps = MMC_MODE_4BIT | MMC_MODE_8BIT;
#if 0
	host->caps |= MMC_CAP_1_8V_DDR | MMC_CAP_UHS_DDR50;
#endif

	hi_mci_sys_init();

	tmp_reg = readl(REG_BASE_CRG + REG_PERI_CRG40);
	switch ((tmp_reg & (0x03 << 8)) >> 8) {
		case 0:
			host->bus_hz = 75000000U;
			break;
		case 1:
			host->bus_hz = 100000000U;
			break;
		case 2:
			host->bus_hz = 50000000U;
			break;
		case 3:
			host->bus_hz = 15000000U;
			break;
	}

	err = add_dwmci(host, host->bus_hz, 400000);
	if (err) {
		printf("s5 mmc init failed...\n");
		return err;
	}

	return 0;
}

int board_mmc_init(bd_t *bis)
{
	return s5_mmc_init();
}

int select_boot_part(struct mmc *mmc, int boot_part)
{
	u8 ext_csd[512];
	unsigned int tmp;
	int err = 0;

	err = mmc_send_ext_csd(mmc, ext_csd);
	if (err) {
		printf("Read boot part error!");
		return -1;
	}

	boot_part = boot_part & 0x7;
	tmp = ext_csd[EXT_CSD_PART_CONF] & 0x7;
	if (tmp == boot_part)
		return 0;

	tmp = ext_csd[EXT_CSD_PART_CONF];
	tmp &= (~0x7);
	tmp |= boot_part;
	err = mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, EXT_CSD_PART_CONF, tmp);
	if (err) {
		printf("Select boot part error!");
		return -1;
	}

	err = mmc_send_ext_csd(mmc, ext_csd);
	if (err) {
		printf("Read boot part error!");
		return -1;
	}

	tmp = ext_csd[EXT_CSD_PART_CONF] & 0x7;
	if (tmp == boot_part)
		printf("Select boot part OK!\n");
	else
		printf("Select boot part Fail!\n");

	return (tmp == boot_part) ? 0 : -1;
}

