/******************************************************************************
 *    NAND Flash Controller V610 Device Driver
 *    Copyright (c) 2009-2010 by Hisilicon.
 *    All rights reserved.
 * ***
 *    Create By Czyong.
 *
******************************************************************************/

#include <common.h>
#include <nand.h>
#include <asm/io.h>
#include <asm/errno.h>
#include <malloc.h>
#include <linux/mtd/nand.h>
#include "hinfc610_gen.h"

#include "hinfc610.h"
#include "hinfc610_read_retry.h"
/*****************************************************************************/

extern void hinfc610_controller_enable(struct hinfc_host *host, int enable);
extern void nand_spl_ids_register(void);

/*****************************************************************************/

static struct hinfc_host host = {
	.chip = NULL,
};
/*****************************************************************************/

#ifdef CONFIG_ARCH_S5
#  include "hinfc610_s5.c"
#endif

/*****************************************************************************/

static unsigned int get_bits(unsigned char byte)
{
	int ix = 0;
	int num = 0;

	if (byte == 0xFF) return 8;
	if (!byte) return 0;

	while (ix++ < 8) {
		if ((byte & 1))
			num++;
		byte = (byte >> 1);
	}
	return num;
}
/*****************************************************************************/

static void nand_register_dump(void)
{
	int ix;
	unsigned int regbase = CONFIG_HINFC610_REG_BASE_ADDRESS;

	printf("Register dump:");
	for (ix = 0; ix <= 0x98; ix += 0x04) {
		if (!(ix & 0x0F))
			printf("\n0x%08X: ", (regbase + ix));
		printf("%08X ", readl(regbase + ix));
	}
	printf("\n");
}
/*****************************************************************************/

static void hinfc610_dma_transfer(struct hinfc_host *host, int todev)
{
	unsigned long reg_val;
	unsigned int dma_addr = (unsigned int)host->dma_buffer;

	hinfc_write(host, dma_addr, HINFC610_DMA_ADDR_DATA);

	dma_addr += HINFC610_DMA_ADDR_OFFSET;
	hinfc_write(host, dma_addr, HINFC610_DMA_ADDR_DATA1);

	dma_addr += HINFC610_DMA_ADDR_OFFSET;
	hinfc_write(host, dma_addr, HINFC610_DMA_ADDR_DATA2);

	dma_addr += HINFC610_DMA_ADDR_OFFSET;
	hinfc_write(host, dma_addr, HINFC610_DMA_ADDR_DATA3);

	dma_addr += HINFC610_DMA_ADDR_OFFSET;
	hinfc_write(host, dma_addr, HINFC610_DMA_ADDR_DATA4);

	dma_addr += HINFC610_DMA_ADDR_OFFSET;
	hinfc_write(host, dma_addr, HINFC610_DMA_ADDR_DATA5);

	dma_addr += HINFC610_DMA_ADDR_OFFSET;
	hinfc_write(host, dma_addr, HINFC610_DMA_ADDR_DATA6);

	dma_addr += HINFC610_DMA_ADDR_OFFSET;
	hinfc_write(host, dma_addr, HINFC610_DMA_ADDR_DATA7);

	hinfc_write(host, host->dma_oob, HINFC610_DMA_ADDR_OOB);

	if (host->ecctype == NAND_ECC_NONE) {
		hinfc_write(host,
			((host->oobsize & HINFC610_DMA_LEN_OOB_MASK)
			 << HINFC610_DMA_LEN_OOB_SHIFT),
			HINFC610_DMA_LEN);

		hinfc_write(host,
			HINFC610_DMA_PARA_DATA_RW_EN
			| HINFC610_DMA_PARA_OOB_RW_EN, 
			HINFC610_DMA_PARA);
	} else
		hinfc_write(host,
			HINFC610_DMA_PARA_DATA_RW_EN
			| HINFC610_DMA_PARA_OOB_RW_EN
			| HINFC610_DMA_PARA_DATA_EDC_EN
			| HINFC610_DMA_PARA_OOB_EDC_EN, 
			HINFC610_DMA_PARA);

	reg_val = (HINFC610_DMA_CTRL_DMA_START
		| HINFC610_DMA_CTRL_BURST4_EN
		| HINFC610_DMA_CTRL_BURST8_EN
		| HINFC610_DMA_CTRL_BURST16_EN
		| ((host->addr_cycle == 4 ? 1 : 0)
		   << HINFC610_DMA_CTRL_ADDR_NUM_SHIFT)
		| ((host->chipselect & HINFC610_DMA_CTRL_CS_MASK)
		   << HINFC610_DMA_CTRL_CS_SHIFT));

	if (todev)
		reg_val |= HINFC610_DMA_CTRL_WE;

	hinfc_write(host, reg_val, HINFC610_DMA_CTRL);

	do {
		unsigned int timeout = 0xF0000000;
		while ((hinfc_read(host, HINFC610_DMA_CTRL))
			& HINFC610_DMA_CTRL_DMA_START && timeout)
			timeout--;
		if (!timeout)
			DBG_BUG("Wait DMA finish timeout.\n");
	} while (0);
}
/*****************************************************************************/

static void hinfc610_cmd_ctrl(struct mtd_info *mtd, int dat, unsigned int ctrl)
{
	int is_cache_invalid = 1;
	struct nand_chip *chip = mtd->priv;
	struct hinfc_host *host = chip->priv;

	if (ctrl & NAND_ALE) {
		unsigned int addr_value = 0;
		unsigned int addr_offset = 0;

		if (ctrl & NAND_CTRL_CHANGE) {
			host->addr_cycle = 0x0;
			host->addr_value[0] = 0x0;
			host->addr_value[1] = 0x0;
		}
		addr_offset = host->addr_cycle << 3;

		if (host->addr_cycle >= HINFC610_ADDR_CYCLE_MASK) {
			addr_offset = (host->addr_cycle -
				HINFC610_ADDR_CYCLE_MASK) << 3;
			addr_value = 1;
		}

		host->addr_value[addr_value] |=
			((dat & 0xff) << addr_offset);

		host->addr_cycle++;
	}

	if ((ctrl & NAND_CLE) && (ctrl & NAND_CTRL_CHANGE)) {
		host->command = dat & 0xff;
		switch (host->command) {
		case NAND_CMD_PAGEPROG:
			host->send_cmd_pageprog(host);
			break;

		case NAND_CMD_READSTART:
			is_cache_invalid = 0;
			host->send_cmd_readstart(host);
			break;

		case NAND_CMD_ERASE2:
			host->send_cmd_erase(host);
			break;

		case NAND_CMD_READID:
			memset((unsigned char *)(chip->IO_ADDR_R), 0, 0x10);
			host->send_cmd_readid(host);
			break;

		case NAND_CMD_STATUS:
			host->send_cmd_status(host);
			break;

		case NAND_CMD_SEQIN:
		case NAND_CMD_ERASE1:
		case NAND_CMD_READ0:
			break;
		case NAND_CMD_RESET:
			host->send_cmd_reset(host, host->chipselect);
			break;

		default :
			break;
		}
	}

	/* pass pagesize and ecctype to kernel when startup. */
	host->enable_ecc_randomizer(host, ENABLE, ENABLE);

	if ((dat == NAND_CMD_NONE) && host->addr_cycle) {
		if (host->command == NAND_CMD_SEQIN ||
		    host->command == NAND_CMD_READ0 ||
		    host->command == NAND_CMD_READID) {
			host->offset = 0x0;
			host->column = (host->addr_value[0] & 0xffff);
		}
	}

	if (is_cache_invalid) {
		host->cache_addr_value[0] = ~0;
		host->cache_addr_value[1] = ~0;
	}
}
/*****************************************************************************/

static int hinfc610_send_cmd_pageprog(struct hinfc_host *host)
{
	if (*host->bbm != 0xFF && *host->bbm != 0x00)
		printf("WARNING: attempt to write an invalid bbm. "
		       "page: 0x%08x, mark: 0x%02x,\n",
		       GET_PAGE_INDEX(host), *host->bbm);

	host->enable_ecc_randomizer(host, ENABLE, ENABLE);

	hinfc_write(host, host->addr_value[0] & 0xffff0000, HINFC610_ADDRL);
	hinfc_write(host, host->addr_value[1], HINFC610_ADDRH);
	hinfc_write(host,
		((NAND_CMD_STATUS << 16) | (NAND_CMD_PAGEPROG << 8) |
		 NAND_CMD_SEQIN),
		HINFC610_CMD);

	*host->epm = host->epmvalue;

	hinfc610_dma_transfer(host, 1);

	return 0;
}
/*****************************************************************************/
#define NAND_BAD_BLOCK              1
#define NAND_EMPTY_PAGE             2
#define NAND_VALID_DATA             3

static int hinfc610_get_data_status(struct hinfc_host *host)
{
	/* this is block start address */
	if (!((host->addr_value[0] >> 16) & host->block_page_mask)) {

		/* it is a bad block */
		if (*host->bbm == 0)
			return NAND_BAD_BLOCK;
		/*
		 * if there are more than 2 bits flipping, it is
		 * maybe a bad block
		 */
		if (host->uc_er && *host->bbm != 0xFF &&
		    get_bits(*host->bbm) <= 6)
			return NAND_BAD_BLOCK;
	}

	/* it is an empty page */
	if (*(host->epm) != host->epmvalue && IS_RANDOMIZER(host))
		return NAND_EMPTY_PAGE;

	return NAND_VALID_DATA;
}

/*****************************************************************************/

static int hinfc610_do_read_retry(struct hinfc_host *host)
{
	int ix;
	unsigned int regval;

	for (ix = 1; host->uc_er && ix < host->read_retry->count; ix++) {

		hinfc_write(host, HINFC610_INTCLR_UE | HINFC610_INTCLR_CE,
			HINFC610_INTCLR);

		host->enable_ecc_randomizer(host, DISABLE, DISABLE);

		host->read_retry->set_rr_param(host, ix);

		/* only enable randomize */
		if (IS_RANDOMIZER(host)) {
			regval = hinfc_read(host, HINFC610_CON);
			regval |= HINFC610_CON_RANDOMIZER_EN;
			hinfc_write(host, regval, HINFC610_CON);
		}

		hinfc_write(host, HINFC610_INTCLR_UE | HINFC610_INTCLR_CE,
			HINFC610_INTCLR);
		hinfc_write(host, host->NFC_CON, HINFC610_CON);
		hinfc_write(host, host->addr_value[0] & 0xffff0000,
			HINFC610_ADDRL);
		hinfc_write(host, host->addr_value[1], HINFC610_ADDRH);
		hinfc_write(host,
			HINFC_CMD_SEQ(NAND_CMD_READ0, NAND_CMD_READSTART),
			HINFC610_CMD);

		hinfc610_dma_transfer(host, 0);
		host->uc_er = (hinfc_read(host, HINFC610_INTS)
			& HINFC610_INTS_UE) ? 1 : 0;
	}

#ifdef CONFIG_NAND_NOTICE_RR
	printf("\nPage %d Do RR(%d/%d) %s.\n", GET_PAGE_INDEX(host),
	       ix-1, host->read_retry->count, (host->uc_er ? "Fail" : "Success"));
#endif

	host->enable_ecc_randomizer(host, DISABLE, DISABLE);

	host->read_retry->reset_rr_param(host);

	return 0;
}
/*****************************************************************************/

static int hinfc610_send_cmd_readstart(struct hinfc_host *host)
{
	if ((host->addr_value[0] == host->cache_addr_value[0])
	    && (host->addr_value[1] == host->cache_addr_value[1]))
		return 0;

	host->enable_ecc_randomizer(host, ENABLE, ENABLE);

	hinfc_write(host, HINFC610_INTCLR_UE | HINFC610_INTCLR_CE,
		HINFC610_INTCLR);
	hinfc_write(host, host->NFC_CON, HINFC610_CON);

	hinfc_write(host, host->addr_value[0] & 0xffff0000, HINFC610_ADDRL);
	hinfc_write(host, host->addr_value[1], HINFC610_ADDRH);
	hinfc_write(host, NAND_CMD_READSTART << 8 | NAND_CMD_READ0,
		HINFC610_CMD);

	hinfc610_dma_transfer(host, 0);
	host->uc_er = (hinfc_read(host, HINFC610_INTS)
		& HINFC610_INTS_UE) ? 1 : 0;

	if (host->read_retry || IS_RANDOMIZER(host)) {
		int status = hinfc610_get_data_status(host);

		if (status == NAND_EMPTY_PAGE) {
			/*
			 * oob area used by yaffs2 only 32 bytes,
			 * so we only fill 32 bytes.
			 */
			if (IS_RANDOMIZER(host))
				memset(host->buffer, 0xFF,
				       host->pagesize + host->oobsize);
			host->uc_er = 2;

		} else if (status == NAND_VALID_DATA) {
			/* if NAND chip support read retry */
			if (host->uc_er && host->read_retry)
				hinfc610_do_read_retry(host);

		} /* 'else' NAND have a bad block, do nothing. */
	}

	host->cache_addr_value[0] = host->addr_value[0];
	host->cache_addr_value[1] = host->addr_value[1];

	return 0;
}
/*****************************************************************************/

static int hinfc610_send_cmd_erase(struct hinfc_host *host)
{
	unsigned int regval;
	/* Don't case the read retry config */
	host->enable_ecc_randomizer(host, DISABLE, DISABLE);

	hinfc_write(host, host->addr_value[0], HINFC610_ADDRL);
	hinfc_write(host, (NAND_CMD_ERASE2 << 8) | NAND_CMD_ERASE1,
		HINFC610_CMD);

	regval = HINFC610_OP_WAIT_READY_EN
		 | HINFC610_OP_CMD2_EN
		 | HINFC610_OP_CMD1_EN
		 | HINFC610_OP_ADDR_EN
		 | ((host->chipselect & HINFC610_OP_NF_CS_MASK)
		    << HINFC610_OP_NF_CS_SHIFT)
		 | ((host->addr_cycle & HINFC610_OP_ADDR_CYCLE_MASK)
		    << HINFC610_OP_ADDR_CYCLE_SHIFT);

	hinfc_write(host, regval, HINFC610_OP);

	WAIT_CONTROLLER_FINISH();

	return 0;
}
/*****************************************************************************/

static int hinfc610_send_cmd_readid(struct hinfc_host *host)
{
	unsigned int regval;

	host->enable_ecc_randomizer(host, DISABLE, DISABLE);
	hinfc_write(host, 0xfffffff, HINFC610_SYNC_TIMING);

	hinfc_write(host, HINFC610_NANDINFO_LEN, HINFC610_DATA_NUM);
	hinfc_write(host, NAND_CMD_READID, HINFC610_CMD);
	hinfc_write(host, 0, HINFC610_ADDRL);

	regval = HINFC610_OP_CMD1_EN
		 | HINFC610_OP_ADDR_EN
		 | HINFC610_OP_READ_DATA_EN
		 | HINFC610_OP_WAIT_READY_EN
		 | ((host->chipselect & HINFC610_OP_NF_CS_MASK)
		    << HINFC610_OP_NF_CS_SHIFT)
		 | (1 << HINFC610_OP_ADDR_CYCLE_SHIFT);

	hinfc_write(host, regval, HINFC610_OP);

	host->addr_cycle = 0x0;

	WAIT_CONTROLLER_FINISH();

	return 0;
}
/*****************************************************************************/

static int hinfc610_enable_ecc_randomizer(struct hinfc_host *host, int ecc_en,
					  int randomizer_en)
{
	unsigned int nfc_con;

	if (IS_RANDOMIZER(host)) {
		if (randomizer_en) {
			host->NFC_CON |= HINFC610_CON_RANDOMIZER_EN;
			host->NFC_CON_ECC_NONE |= HINFC610_CON_RANDOMIZER_EN;
		} else {
			host->NFC_CON &= ~HINFC610_CON_RANDOMIZER_EN;
			host->NFC_CON_ECC_NONE &= ~HINFC610_CON_RANDOMIZER_EN;
		}
	}

	nfc_con = (ecc_en ? host->NFC_CON : host->NFC_CON_ECC_NONE);

	hinfc_write(host, nfc_con, HINFC610_CON);

	return 0;
}
/*****************************************************************************/

static int hinfc610_send_cmd_status(struct hinfc_host *host)
{
	unsigned int regval;

	host->enable_ecc_randomizer(host, DISABLE, DISABLE);

	hinfc_write(host, HINFC610_NANDINFO_LEN, HINFC610_DATA_NUM);
	hinfc_write(host, NAND_CMD_STATUS, HINFC610_CMD);

	regval = HINFC610_OP_CMD1_EN
		 | HINFC610_OP_READ_DATA_EN
		 | HINFC610_OP_WAIT_READY_EN
		 | ((host->chipselect & HINFC610_OP_NF_CS_MASK)
		    << HINFC610_OP_NF_CS_SHIFT);

 	hinfc_write(host, regval, HINFC610_OP);

	WAIT_CONTROLLER_FINISH();

	return 0;
}

/*****************************************************************************/
static int hinfc610_send_cmd_reset(struct hinfc_host *host, int chipselect)
{
	unsigned int regval;

	hinfc_write(host, NAND_CMD_RESET, HINFC610_CMD);

	regval = HINFC610_OP_CMD1_EN
		 | (((chipselect & HINFC610_OP_NF_CS_MASK)
		    << HINFC610_OP_NF_CS_SHIFT)
		 | HINFC610_OP_WAIT_READY_EN);

	hinfc_write(host, regval, HINFC610_OP);

	WAIT_CONTROLLER_FINISH();

	return 0;
}
/*****************************************************************************/

static int hinfc610_dev_ready(struct mtd_info *mtd)
{
	return 0x1;
}
/*****************************************************************************/

static void hinfc610_select_chip(struct mtd_info *mtd, int chipselect)
{
	struct nand_chip *chip = mtd->priv;
	struct hinfc_host *host = chip->priv;

	if (chipselect < 0)
		return;

	if (chipselect > CONFIG_HINFC610_MAX_CHIP)
		DBG_BUG("invalid chipselect: %d\n", chipselect);

	host->chipselect = chipselect;
}
/*****************************************************************************/

static uint8_t hinfc610_read_byte(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
	struct hinfc_host *host = chip->priv;

	if (host->command == NAND_CMD_STATUS)
		return readb(chip->IO_ADDR_R);

	host->offset++;

	if (host->command == NAND_CMD_READID)
		return readb(chip->IO_ADDR_R + host->offset - 1);

	return readb(host->buffer + host->column + host->offset - 1);
}
/*****************************************************************************/

static u16 hinfc610_read_word(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
	struct hinfc_host *host = chip->priv;

	host->offset += 2;
	return readw(host->buffer + host->column + host->offset - 2);
}
/*****************************************************************************/

static void hinfc610_write_buf(struct mtd_info *mtd,
	const uint8_t *buf, int len)
{
	struct nand_chip *chip = mtd->priv;
	struct hinfc_host *host = chip->priv;

	memcpy(host->buffer + host->column + host->offset, buf, len);
	host->offset += len;
}
/*****************************************************************************/

static void hinfc610_read_buf(struct mtd_info *mtd, uint8_t *buf, int len)
{
	struct nand_chip *chip = mtd->priv;
	struct hinfc_host *host = chip->priv;

	memcpy(buf, host->buffer + host->column + host->offset, len);
	host->offset += len;
}
/*****************************************************************************/

static struct nand_ecclayout nand_ecc_default =
{
	.oobfree = {{2, 30}}
};
/*****************************************************************************/

static struct nand_ecclayout nand_ecc_2k_13bit =
{
	.oobfree = {{2, 14}}
};
/*****************************************************************************/
/* ecc/pagesize get from CPU PIN. */
static struct nand_config_table hinfc610_hw_pin_config_table[] =
{
	{NAND_PAGE_32K, NAND_ECC_60BIT, 3520/*3520*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_41BIT, 2368/*2368*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_NONE,  32 ,          &nand_ecc_default},

	{NAND_PAGE_16K, NAND_ECC_80BIT, 1376/*1376*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_72BIT, 1376/*1376*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_41BIT, 1184/*1184*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_NONE,  32 ,          &nand_ecc_default},

	{NAND_PAGE_8K, NAND_ECC_80BIT, 1152/*1152*/, &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_72BIT, 1040/*1040*/, &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_60BIT, 896 /*896*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_41BIT, 608 /*608*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_24BIT, 368 /*368*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_NONE,  32,           &nand_ecc_default},

	{NAND_PAGE_4K, NAND_ECC_27BIT, 224 /*224*/,  &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_24BIT, 200 /*200*/,  &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_18BIT, 160 /*160*/,  &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_13BIT, 112 /*112*/,  &nand_ecc_2k_13bit},
	{NAND_PAGE_4K, NAND_ECC_8BIT,  88  /*88*/,   &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_NONE,  32,           &nand_ecc_default},

	{NAND_PAGE_2K, NAND_ECC_27BIT, 128 /*128*/, &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_24BIT, 116 /*116*/, &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_18BIT, 96  /*96*/,  &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_13BIT, 64  /*64*/,  &nand_ecc_2k_13bit},
	{NAND_PAGE_2K, NAND_ECC_8BIT,  60  /*60*/,  &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_NONE,  32,          &nand_ecc_default},

	{0,0,0,NULL},
};

/*****************************************************************************/
/* ecc/pagesize get from NAND controller */
static struct nand_config_table hinfc610_hw_auto_config_table[] =
{
	{NAND_PAGE_32K, NAND_ECC_80BIT, 4544/*4544*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_60BIT, 3520/*3520*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_41BIT, 2368/*2368*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_NONE,  32 ,          &nand_ecc_default},

	{NAND_PAGE_16K, NAND_ECC_80BIT, 1376/*1376*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_60BIT, 1760/*1760*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_41BIT, 1184/*1184*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_NONE,  32 ,          &nand_ecc_default},

	{NAND_PAGE_8K, NAND_ECC_80BIT, 1152/*1152*/, &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_60BIT, 896 /*896*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_41BIT, 608 /*608*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_24BIT, 368 /*368*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_NONE,  32,           &nand_ecc_default},

	{NAND_PAGE_4K, NAND_ECC_24BIT, 200 /*200*/,  &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_8BIT,  88 /*88*/,    &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_NONE,  32,           &nand_ecc_default},

	{NAND_PAGE_2K, NAND_ECC_24BIT, 116 /*116*/, &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_8BIT,  60  /*60*/,  &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_NONE,  32,          &nand_ecc_default},

	{0,0,0,NULL},
};
/*****************************************************************************/
/* ecc/pagesize get from NAND id */
static struct nand_config_table hinfc610_soft_auto_config_table[] =
{
	{NAND_PAGE_32K, NAND_ECC_80BIT, 4544/*4544*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_72BIT, 4096/*4096*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_60BIT, 3520/*3520*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_48BIT, 2752/*2752*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_41BIT, 2368/*2368*/, &nand_ecc_default},
	{NAND_PAGE_32K, NAND_ECC_NONE,  32 ,          &nand_ecc_default},

	{NAND_PAGE_16K, NAND_ECC_80BIT, 1376/*1376*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_72BIT, 1376/*1376*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_60BIT, 1760/*1760*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_48BIT, 1376/*1376*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_41BIT, 1184/*1184*/, &nand_ecc_default},
	{NAND_PAGE_16K, NAND_ECC_NONE,  32 ,          &nand_ecc_default},

	{NAND_PAGE_8K, NAND_ECC_80BIT, 1152/*1152*/, &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_72BIT, 1040/*1040*/, &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_60BIT, 896 /*896*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_48BIT, 704 /*704*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_41BIT, 608 /*608*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_32BIT, 480 /*480*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_27BIT, 416 /*416*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_24BIT, 368 /*368*/,  &nand_ecc_default},
	{NAND_PAGE_8K, NAND_ECC_NONE,  32,           &nand_ecc_default},

	{NAND_PAGE_4K, NAND_ECC_32BIT, 256 /*256*/,  &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_27BIT, 224 /*224*/,  &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_24BIT, 200 /*200*/,  &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_18BIT, 160 /*160*/,  &nand_ecc_default},
#ifndef CONFIG_SUPPORT_YAFFS
	{NAND_PAGE_4K, NAND_ECC_13BIT, 112 /*112*/,  &nand_ecc_2k_13bit},
#endif
	{NAND_PAGE_4K, NAND_ECC_8BIT,  88  /*88*/,   &nand_ecc_default},
	{NAND_PAGE_4K, NAND_ECC_NONE,  32,           &nand_ecc_default},

#ifndef CONFIG_SUPPORT_YAFFS
	{NAND_PAGE_2K, NAND_ECC_32BIT, 128 /*128*/, &nand_ecc_2k_13bit},
#endif
	{NAND_PAGE_2K, NAND_ECC_27BIT, 128 /*128*/, &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_24BIT, 116 /*116*/, &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_18BIT, 96  /*96*/,  &nand_ecc_default},
#ifndef CONFIG_SUPPORT_YAFFS
	{NAND_PAGE_2K, NAND_ECC_13BIT, 64  /*64*/,  &nand_ecc_2k_13bit},
#endif
	{NAND_PAGE_2K, NAND_ECC_8BIT,  60  /*60*/,  &nand_ecc_default},
	{NAND_PAGE_2K, NAND_ECC_NONE,  32,          &nand_ecc_default},

	{0,0,0,NULL},

};
/*****************************************************************************/

static struct nand_config_table *
hinfc610_get_best_ecc(struct nand_config_table *config, struct mtd_info *mtd)
{
	struct nand_config_table *best = NULL;

	for (; config->layout; config++) {
		if (nandpage_type2size(config->pagetype) != mtd->writesize)
			continue;

		if (mtd->oobsize < config->oobsize)
			continue;

		if (!best || (best->ecctype < config->ecctype))
			best = config;
	}

	if (!best)
		DBG_BUG(ERSTR_DRIVER
		        "Driver does not support the pagesize(%d) "
		        "and oobsize(%d).\n",
		        mtd->writesize, mtd->oobsize);

	return best;
}
/*****************************************************************************/
/* force the pagesize and ecctype */
struct nand_config_table *
hinfc610_force_ecc(struct nand_config_table *config, struct mtd_info *mtd,
		   int pagetype, int ecctype, char *cfgmsg, int allow_pagediv)
{
	int pagesize;
	struct nand_config_table *fit = NULL;

	for (; config->layout; config++) {
		if (config->pagetype == pagetype &&
		    config->ecctype == ecctype) {
			fit = config;
			break;
		}
	}

	if (!fit) {
		nand_register_dump();
		DBG_BUG(ERSTR_DRIVER
		        "Driver(%s mode) does not support this Nand Flash "
		        "pagesize:%s, ecctype:%s\n",
		        cfgmsg,
		        nand_page_name(pagetype),
		        nand_ecc_name(ecctype));
		return NULL;
	}

	pagesize = nandpage_type2size(pagetype);
	if ((pagesize != mtd->writesize) &&
	    (pagesize > mtd->writesize || !allow_pagediv)) {
		nand_register_dump();
		DBG_BUG(ERSTR_HARDWARE
		        "Hardware (%s mode) configure pagesize %d, "
		        "but the Nand Flash pageszie is %d\n",
		        cfgmsg,
		        pagesize,
		        mtd->writesize);
		return NULL;
	}

	if (fit->oobsize > mtd->oobsize) {
		DBG_BUG(ERSTR_HARDWARE
		        "(%s mode) The Nand Flash offer space area is %d bytes, "
		        "but the controller request %d bytes in ecc %s. "
		        "Please make sure the hardware ECC configuration is correct.",
		        cfgmsg, mtd->oobsize, fit->oobsize,
		        nand_ecc_name(ecctype));
		return NULL;
	}

	return fit;
}
/*****************************************************************************/

static int hinfc610_check_config(struct hinfc_host *host,
				 struct nand_dev_t *nand_dev,
				 struct nand_config_table *best, int is_hw_auto)
{
	int isdump = 0;
	int ecctype;
	int pagetype;

	pagetype = hinfc610_page_reg2type(
		(host->NFC_CON >> HINFC610_CON_PAGEISZE_SHIFT) &
		 HINFC610_CON_PAGESIZE_MASK);

	ecctype = hinfc610_ecc_reg2type(
		(host->NFC_CON >> HINFC610_CON_ECCTYPE_SHIFT) &
		 HINFC610_CON_ECCTYPE_MASK);

	if (pagetype != best->pagetype) {
		if (!isdump++)
			nand_register_dump();
		if (is_hw_auto) {
			printk("!!! warning: "
			       "Hardware configure pagesize %s, "
			       "but the Nand Flash pageszie is %s. "
			       "Update fastboot will resolve this problem.\n",
			       nand_page_name(pagetype),
			       nand_page_name(best->pagetype));
		} else {
			printk("!!! warning: "
			       "Hardware configure pagesize %s, "
			       "but the Nand Flash pageszie is %s. "
			       "Modify hardware pin will resolve this problem.\n",
			       nand_page_name(pagetype),
			       nand_page_name(best->pagetype));
		}
	}

	if (ecctype != best->ecctype) {
		if (!isdump++)
			nand_register_dump();
		if (is_hw_auto) {
			DBG_BUG("Hardware configure ecctype %s, "
				"but %s is more suitable for this Nand Flash. "
				"Update fastboot will resolve this problem.\n",
				nand_ecc_name(ecctype),
				nand_ecc_name(best->ecctype));
		} else {
			printk("!!! warning: "
				"Hardware configure ecctype %s, "
				"but %s is more suitable for this Nand Flash.\n",
				nand_ecc_name(ecctype),
				nand_ecc_name(best->ecctype));
			best->ecctype = ecctype;
		}
	}

	if (IS_RANDOMIZER(nand_dev) && !IS_RANDOMIZER(host)) {
		if (is_hw_auto) {
			DBG_BUG("Hardware is not configure randomizer, "
				"but it is more suitable for this Nand Flash. "
				"Update fastboot will resolve this problem.\n");
		} else {
			DBG_BUG("Hardware is not configure randomizer, "
				"but it is more suitable for this Nand Flash. "
				"Modify hardware pin will resolve this problem.\n");
		}
	}

	return 0;
}
/*****************************************************************************/

static int hinfc610_ecc_probe(struct mtd_info *mtd, struct nand_chip *chip,
			      struct nand_dev_t *nand_dev)
{
	char *start_type = "unknown";
	struct nand_config_table *best = NULL;
	struct nand_config_table *config = NULL;
	struct hinfc_host *host = chip->priv;

	if (get_bootmedia(NULL, NULL) == BOOT_MEDIA_NAND) {
		if (IS_HW_AUTO(host)) {
			config = hinfc610_hw_auto_config_table;
			start_type = "HW-Auto";
			best = hinfc610_get_best_ecc(config, mtd);
			if (IS_CONFIG_DONE(host)) {
				hinfc610_check_config(host, nand_dev, best, 1);
				nand_dev->flags |= host->flags;
			}
		} else {
			config = hinfc610_hw_pin_config_table;
			start_type = "HW-Pin";
			best = hinfc610_get_best_ecc(config, mtd);
			hinfc610_check_config(host, nand_dev, best, 0);
			nand_dev->flags |= host->flags;
		}
	} else {
		config = hinfc610_soft_auto_config_table;
		start_type = "Soft-Auto";
		best = hinfc610_get_best_ecc(config, mtd);
	}
	host->flags = nand_dev->flags;

#ifdef CONFIG_HINFC610_PAGESIZE_AUTO_ECC_NONE
	{
		int pagetype = nandpage_size2type(mtd->writesize);
		int ecctype = NAND_ECC_NONE;
		best = hinfc610_force_ecc(config, mtd, pagetype, ecctype,
			"force config", 0);
		start_type = "NoECC";
	}
#endif /* CONFIG_HINFC610_PAGESIZE_AUTO_ECC_NONE */

	if (!best)
		DBG_BUG(ERSTR_HARDWARE
		        "Please configure Nand Flash pagesize and ecctype!\n");

	if (best->ecctype != NAND_ECC_NONE)
		mtd->oobsize = best->oobsize;
	mtd->oobused = HINFC610_OOBSIZE_FOR_YAFFS;
	chip->ecc.layout = best->layout;

	host->ecctype  = best->ecctype;
	host->pagesize = nandpage_type2size(best->pagetype);
	host->oobsize  = mtd->oobsize;
	host->block_page_mask = ((mtd->erasesize / mtd->writesize) - 1);

	host->buforg = kmalloc((host->pagesize
		+ host->oobsize + HINFC610_DMA_ALIGN), GFP_KERNEL);
	if (!host->buforg) {
		DBG_BUG(ERSTR_DRIVER
			"Can't malloc memory for NAND driver.");
		return 1;
	}
	/* 32 bytes alignment */
	host->buffer = (char *)((unsigned int)(host->buforg
		+ HINFC610_DMA_ALIGN - 1) & ~(HINFC610_DMA_ALIGN - 1));
	host->dma_buffer = (unsigned int)host->buffer;

	memset((char *)chip->IO_ADDR_R, 0xff,
	       HINFC610_BUFFER_BASE_ADDRESS_LEN);
	memset(host->buffer, 0xff,
		(host->pagesize + host->oobsize));

	host->dma_oob = host->dma_buffer + host->pagesize;
	host->bbm = (unsigned char *)(host->buffer
		+ host->pagesize + HINFC_BAD_BLOCK_POS);

	host->epm = (unsigned short *)(host->buffer
		+ host->pagesize + chip->ecc.layout->oobfree[0].offset + 28);

	host->NFC_CON  = (HINFC610_CON_OP_MODE_NORMAL
		| ((hinfc610_page_type2reg(best->pagetype) &
		    HINFC610_CON_PAGESIZE_MASK)
		   << HINFC610_CON_PAGEISZE_SHIFT)
		| HINFC610_CON_READY_BUSY_SEL
		| ((hinfc610_ecc_type2reg(best->ecctype) &
		    HINFC610_CON_ECCTYPE_MASK)
		   << HINFC610_CON_ECCTYPE_SHIFT));

	host->NFC_CON_ECC_NONE = (HINFC610_CON_OP_MODE_NORMAL
		| ((hinfc610_page_type2reg(best->pagetype) &
		    HINFC610_CON_PAGESIZE_MASK)
		   << HINFC610_CON_PAGEISZE_SHIFT)
		| HINFC610_CON_READY_BUSY_SEL);

	if (mtd->writesize > NAND_MAX_PAGESIZE ||
	    mtd->oobsize > NAND_MAX_OOBSIZE) {
		DBG_BUG(ERSTR_DRIVER
		        "Driver does not support this Nand Flash. "
		        "Please increase NAND_MAX_PAGESIZE and NAND_MAX_OOBSIZE.\n");
	}

	if (mtd->writesize != host->pagesize) {
		unsigned int shift = 0;
		unsigned int writesize = mtd->writesize;
		while (writesize > host->pagesize) {
			writesize >>= 1;
			shift++;
		}
		chip->chipsize = chip->chipsize >> shift;
		mtd->erasesize = mtd->erasesize >> shift;
		mtd->writesize = host->pagesize;
		printk("Nand divide into 1/%u\n", (1 << shift));
	}

	nand_dev->start_type = start_type;
	nand_dev->ecctype = host->ecctype;

	host->read_retry = NULL;
	if (nand_dev->read_retry_type != NAND_RR_NONE) {
		host->read_retry
			= hinfc610_find_read_retry(nand_dev->read_retry_type);
		if (!host->read_retry) {
			DBG_BUG(ERSTR_DRIVER
			        "This Nand Flash need to enable the "
			        "'read retry' feature. "
			        "but the driver dose not offer the feature");
		}
	}

	/*
	 * If it want to support the 'read retry' feature, the 'randomizer'
	 * feature must be support first.
	 */
	if (host->read_retry && !IS_RANDOMIZER(host)) {
		DBG_BUG(ERSTR_HARDWARE
		        "This Nand flash need to enable 'randomizer' "
		        "feature. Please configure hardware randomizer PIN.");
	}

	/*
	 * Check if hardware enable randomizer PIN, But NAND does not need
	 * randomizer. We will notice user.
	 */
	while (IS_RANDOMIZER(host)) {
		if (host->pagesize == _8K
		    && (host->ecctype >= NAND_ECC_24BIT
		    && host->ecctype <= NAND_ECC_80BIT))
			break;

		if (host->pagesize == _16K
		    && (host->ecctype >= NAND_ECC_41BIT
		    && host->ecctype <= NAND_ECC_80BIT))
			break;

		if (host->pagesize == _32K
		    && (host->ecctype >= NAND_ECC_41BIT
		    && host->ecctype <= NAND_ECC_80BIT))
			break;

		DBG_BUG(ERSTR_HARDWARE
		        "This NAND flash does not support `randomizer`, "
		        "Please don't configure hardware randomizer PIN.");
	}

	nand_dev->hostver = HINFC_VER_610;

	return 0;
}
/*****************************************************************************/

int hinfc610_nand_get_rr_param(char *param)
{
	int ret = 0;

	if ((host.read_retry->type == NAND_RR_HYNIX_BG_CDIE) ||
	    (host.read_retry->type == NAND_RR_HYNIX_CG_ADIE)) {
		ret = 64;
		memcpy(param, host.rr_data, ret);
	}
	return ret;
}
/*****************************************************************************/

static int hinfc610_nand_init(struct hinfc_host *host, struct nand_chip *chip)
{
	unsigned int regval;

	chip->priv        = host;
	chip->cmd_ctrl    = hinfc610_cmd_ctrl;
	chip->dev_ready   = hinfc610_dev_ready;
	chip->select_chip = hinfc610_select_chip;
	chip->read_byte   = hinfc610_read_byte;
	chip->read_word   = hinfc610_read_word;
	chip->write_buf   = hinfc610_write_buf;
	chip->read_buf    = hinfc610_read_buf;

	chip->chip_delay = HINFC610_CHIP_DELAY;
	chip->options    = NAND_NO_AUTOINCR | NAND_BBT_SCANNED;

	chip->ecc.layout = NULL;
	chip->ecc.mode   = NAND_ECC_NONE;

	host->chip          = chip;
	host->addr_cycle    = 0;
	host->addr_value[0] = 0;
	host->addr_value[1] = 0;
	host->cache_addr_value[0] = ~0;
	host->cache_addr_value[1] = ~0;
	host->chipselect    = 0;

	host->send_cmd_pageprog  = hinfc610_send_cmd_pageprog;
	host->send_cmd_readstart = hinfc610_send_cmd_readstart;
	host->send_cmd_erase     = hinfc610_send_cmd_erase;
	host->send_cmd_readid    = hinfc610_send_cmd_readid;
	host->send_cmd_status    = hinfc610_send_cmd_status;
	host->send_cmd_reset     = hinfc610_send_cmd_reset;

	regval = hinfc_read(host, HINFC610_CON);

	if (regval & HINFC610_CON_RANDOMIZER_EN)
		host->flags |= NAND_RANDOMIZER;

	host->NFC_CON = (regval
		| HINFC610_CON_OP_MODE_NORMAL
		| HINFC610_CON_READY_BUSY_SEL);

	host->NFC_CON_ECC_NONE = (host->NFC_CON
		& (~(HINFC610_CON_ECCTYPE_MASK
		     << HINFC610_CON_ECCTYPE_SHIFT))
		& (~HINFC610_CON_RANDOMIZER_EN));

	hinfc_write(host,
		(SET_HINFC610_PWIDTH(CONFIG_HINFC610_W_LATCH,
		 CONFIG_HINFC610_R_LATCH,
		 CONFIG_HINFC610_RW_LATCH)),
		HINFC610_PWIDTH);

	host->version = hinfc_read(host, HINFC610_VERSION);

	regval = hinfc_read(host, HINFC610_BOOT_CFG);
	if (regval & HINFC610_BOOT_CFG_SAVE_PIN_MODE)
		host->flags |= NAND_HW_AUTO;

	if (regval & HINFC610_BOOT_CFG_SYC_NAND_PAD)
		host->flags |= NAND_SYNCHRONOUS_BOOT;

	/* check if start form nand or not */
	if (readl(REG_BASE_SCTL + REG_SC_GEN15) == BOOT_FROM_NAND)
		host->flags |= NAND_CONFIG_DONE;

	host->enable_ecc_randomizer = hinfc610_enable_ecc_randomizer;
	host->epmvalue = 0x0000;

	nand_param_adjust = hinfc610_ecc_probe;

	return 0;
}
/*****************************************************************************/

static int hinfc610_version_check(void)
{
	unsigned long regval;
	regval = readl(CONFIG_HINFC610_REG_BASE_ADDRESS + HINFC610_VERSION);
	return (regval == HINFC_VER_610);
}
/*****************************************************************************/

int hinfc610_init(struct nand_chip *chip)
{
	printf("Check nand flash controller v610. ");
	if (!hinfc610_version_check()) {
		printf("\n");
		return -ENODEV;
	}
	printf("found\n");

	if (host.chip)
		return 0;

#ifdef CONFIG_NAND_NOTICE_RR
	printf("Nand CONFIG_NAND_NOTICE_RR is defined\n");
#endif /* CONFIG_NAND_NOTICE_RR */

	memset((char *)&host, 0, sizeof(struct hinfc_host));

	host.iobase = (void __iomem *)CONFIG_HINFC610_REG_BASE_ADDRESS;
	chip->IO_ADDR_R = chip->IO_ADDR_W =
		(void *)CONFIG_HINFC610_BUFFER_BASE_ADDRESS;

	hinfc610_controller_enable(&host, 1);

	if (hinfc610_nand_init(&host, chip)) {
		printk("failed to allocate device buffer.\n");
		return -ENOMEM;
	}

	nand_spl_ids_register();

	return 0;
}
/*****************************************************************************/

#include <nand_logif.h>
#include <spiflash_logif.h>
static unsigned char nand_otp_buf[0x4000];

int hinfc610_nand_rr_param_init(void)
{
	if (!host.read_retry)
		return 0;

	if (host.read_retry->type == NAND_RR_HYNIX_BG_CDIE
	    || host.read_retry->type == NAND_RR_HYNIX_CG_ADIE) {
#define NAND_RR_PARAM_TAG       0xB31E1F8C

		int ix, ret;
		unsigned char *ss = nand_otp_buf;

		/*
		 * do something(ex:read) before reading nand otp,
		 * otherwise hynix maybe wrong.
		 */
		struct nand_logic_t *nand_logic
			= nand_logic_open(0ULL, 0x200000ULL);
		if (!nand_logic) {
			printf("Error: Can't open nand flash. "
				"Init otp parameters failed.\n");
			return -1;
		}
		ret = nand_logic_read(nand_logic, 0x0,
				0x4000, nand_otp_buf, 0);
		if (ret <= 0) {
			printf("Error: Can't read nand flash. "
				"Init otp parameters failed.\n");
			return -1;
		}
		nand_logic_close(nand_logic);

		struct spiflash_logic_t *spiflash_logic = spiflash_logic_open(0,
			0x100000ULL);
		if (!spiflash_logic) {
			printf("Error: Can't open spi flash. "
				"Init otp parameters failed.\n");
			return -1;
		}

		printf("Save Nand read retry paramter to SPIFlash offset 0x%llx\n",
			NAND_RR_PARAM_OFFSET);

		ret = spiflash_logic_read(spiflash_logic, NAND_RR_PARAM_OFFSET,
				8192, nand_otp_buf);
		if (ret != 8192) {
			printf("Error: Can't read spi flash. "
				"Init otp parameters failed.\n");
			return -1;
		}

		if ((*(unsigned int *)ss) == NAND_RR_PARAM_TAG) {
			/* get parameter in NAND/SPIFlash */
			unsigned char chr;

			for (ix = 0, ss += sizeof(int); ix < 64 && *ss; ix++) {

				chr = *ss - ((*ss >= 'A') ? ('A' - 10) : '0');
				host.rr_data[ix] = ((chr << 4) & 0xF0);
				ss++;
				chr = *ss - ((*ss >= 'A') ? ('A' - 10) : '0');
				host.rr_data[ix] |= (chr & 0xF);
				ss++;
				ss++;
			}

		} else {
			/* Get RR parameter from NAND OTP table. */
			if (host.read_retry->get_rr_param(&host)) {
				spiflash_logic_close(spiflash_logic);
				return -1;
			}

			*(int *)ss = NAND_RR_PARAM_TAG;
			ss += sizeof(int);

			for (ix = 0; ix < 64; ix++)
				ss += sprintf((char *)ss, "%02X ",
					(unsigned int)host.rr_data[ix]);

			ret = spiflash_logic_write(spiflash_logic, NAND_RR_PARAM_OFFSET,
				8192, nand_otp_buf);
			if (ret != 8192) {
				printf("Error: Can't write spi flash. "
					"Init otp parameters failed.\n");
				return -1;
			}
		}

		spiflash_logic_close(spiflash_logic);

#undef NAND_RR_PARAM_TAG
	}

	return 0;
}
