/*XHCI for Hi3535*/
#define HIUSB_XHCI_BASE        0xF9870000

#define PERI_CRG123            (0xF8A22000 + 0x1ec)
#define USB3_VCC_SRST_REQ2     (1<<12)
#define USB3_PHY_SRST_REQ2       (0x1 << 13)
#define USB3_PHY_SRST_TREQ2     (0x1 << 14)

#define GTXTHRCFG             0xc108
#define GRXTHRCFG             0xc10c
#define REG_GCTL              0xc110

#define REG_GUSB2PHYCFG0         0xC200
#define BIT_UTMI_ULPI         (0x1 << 4)
#define BIT_UTMI_8_16         (0x1 << 3)

#define REG_GUSB3PIPECTL0	0xc2c0
#define PCS_SSP_SOFT_RESET	(0x1 << 31)


int xhci_hcd_init(int index, struct xhci_hccr **hccr, struct xhci_hcor **hcor)
{
	*hccr = (struct xhci_hccr *)(HIUSB_XHCI_BASE);
	*hcor = (struct xhci_hcor *)((uint32_t) *hccr
			+ HC_LENGTH(xhci_readl(&(*hccr)->cr_capbase)));

	return 0;
}

static void hiusb_start_hcd(void)
{
	unsigned int reg;


//	reg = readl(IO_ADDRESS(0xf8a220bc));
	writel(0x300007, (0xf8a220bc));
	/* de-assert usb3_vcc_srst_req */
//	writel(0xFF4F000, PERI_CRG123);
	writel(0x7f000, PERI_CRG123);
	mdelay(100);
	reg = readl(PERI_CRG123);
	reg &= ~(USB3_VCC_SRST_REQ2);
	writel(reg, PERI_CRG123);
	mdelay(100);

	reg = readl(HIUSB_XHCI_BASE + REG_GUSB3PIPECTL0);
	reg |= PCS_SSP_SOFT_RESET;
	writel(reg, HIUSB_XHCI_BASE + REG_GUSB3PIPECTL0);

#ifdef CONFIG_S5_FPGA
	/* option1: configure for TMT, utmi 16bit interface */
	reg = readl(HIUSB_XHCI_BASE + REG_GUSB2PHYCFG0);
	reg &= ~(0xf<<10);
	reg |= (0x9<<10);
	reg &= ~BIT_UTMI_8_16;
	reg &= ~(1<<6);       //disable suspend
	writel(reg, HIUSB_XHCI_BASE + REG_GUSB2PHYCFG0);
	mdelay(20);
#else
	/*step 3: USB2 PHY chose ulpi 8bit interface */
	reg = readl(HIUSB_XHCI_BASE + REG_GUSB2PHYCFG0);
	reg &= ~BIT_UTMI_ULPI;
	reg &= ~(BIT_UTMI_8_16);
	reg &= ~(1<<6);  //disable phy suspend
	writel(reg, HIUSB_XHCI_BASE + REG_GUSB2PHYCFG0);
	mdelay(20);
#endif

	reg = readl(HIUSB_XHCI_BASE + REG_GCTL);
	reg &= ~(0x3<<12);
	reg |= (0x1<<12); /*[13:12] 01: Host; 10: Device; 11: OTG*/
	writel(reg, HIUSB_XHCI_BASE + REG_GCTL);

	/* de-assert usb3phy hard-macro por */
	reg = readl(PERI_CRG123);
	reg &= ~USB3_PHY_SRST_REQ2;
	writel(reg, PERI_CRG123);
	mdelay(100);

/*	reg = readl(PERI_CRG123);
	reg &= ~USB3_PHY_SRST_TREQ2; // de-assert hsp hard-macro port reset
	writel(reg, PERI_CRG123);*/
//	printk("\n#######%s,%d,readl(0xf9870420)=0x%x,readl(PERI_CRG123)=0x%x\n",__func__,__LINE__,readl(IO_ADDRESS(0xf9870420)),readl(PERI_CRG123));

	reg = readl(HIUSB_XHCI_BASE + REG_GUSB3PIPECTL0);
	reg &= ~PCS_SSP_SOFT_RESET;
	reg &= ~(1<<17);       //disable suspend
	writel(reg, HIUSB_XHCI_BASE + REG_GUSB3PIPECTL0);
	mdelay(100);

	writel(0x23100000, HIUSB_XHCI_BASE + GTXTHRCFG);
	writel(0x23100000, HIUSB_XHCI_BASE + GRXTHRCFG);
	mdelay(20);

	writel(0x022000b0, (0xf8a208d4));
	writel(0x022100b0, (0xf8a208d4));
	writel(0x022000b0, (0xf8a208d4));
	mdelay(20);

    /*fix reset,add control timeout;*/
	writel(0x02008062, (0xf987c12c));
	mdelay(1);
//	writel(0x03f0ff, (0xf8a220b8));
//	mdelay(1);

}

static void hiusb_stop_hcd(void)
{
	unsigned int reg;

	reg = readl(PERI_CRG123);
	writel(reg | (USB3_VCC_SRST_REQ2), PERI_CRG123);
	mdelay(500);

}


