/******************************************************************************
 *    COPYRIGHT (C) 2013 Czyong. Hisilicon
 *    All rights reserved.
 * ***
 *    Create by Czyong 2013-11-04
 *
******************************************************************************/

#include <common.h>
#include <bootimg.h>
#include <linux/err.h>

int check_bootimg(char *buf, unsigned int len)
{
	struct boot_img_hdr *hdr = (struct boot_img_hdr *)buf;

	if (len < sizeof(struct boot_img_hdr))
		return -EINVAL;

	if (memcmp(hdr->magic, BOOT_MAGIC, BOOT_MAGIC_SIZE))
		return -ENODATA;

	return 0;
}

int bootimg_get_kernel(char *buf, unsigned int len, unsigned int *entry)
{
	int ret;
	struct boot_img_hdr *hdr = (struct boot_img_hdr *)buf;

	ret = check_bootimg(buf, len);
	if (ret)
		return ret;

	if (hdr->page_size > len)
		return -EINVAL;

	if (entry)
		*entry = (unsigned int *)((char *)hdr + hdr->page_size);

	return 0;
}

int bootimg_get_ramfs(char *buf, unsigned int len, unsigned int *initrd_start,
		      unsigned int *initrd_size)
{
	int ret;
	struct boot_img_hdr *hdr = (struct boot_img_hdr *)buf;

	ret = check_bootimg(buf, len);
	if (ret)
		return ret;

	if (!hdr->ramdisk_size || (hdr->kernel_size + hdr->page_size) > len)
		return -EINVAL;

	if (initrd_start) {
		unsigned int tmp;

		tmp = ((hdr->page_size << 1) + hdr->kernel_size)
			& ~(hdr->page_size - 1);

		*initrd_start = (unsigned int)hdr + tmp;
	}

	if (initrd_size)
		*initrd_size = hdr->ramdisk_size;

	return 0;
}
